#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 16 19:15:19 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm
from glob import glob
from pprint import pprint

#::: optional modules
try:
    from wotan import flatten
except ImportError:
    pass


#::: my modules
import allesfitter
from allesfitter.io import read_csv
from allesfitter import tessplot
from allesfitter.time_series import sigma_clip, slide_clip, binning, mask_regions
from allesfitter.detection.periodicity import estimate_period
from allesfitter.detection.transit_search import get_tls_kwargs_by_tic, tls_search
from exoworlds.tess import tessio

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




#::: 0) tic_id
tic_id = '390651552'

#::: 1) load data
time, flux, flux_err = tessio.get(tic_id, unpack=True)
tessplot(time, flux, clip=False)
tessplot(time, flux)

#::: 2) prepare data
flux = sigma_clip(time, flux, high=3, low=20)
flux = slide_clip(time, flux, high=3, low=20)
flux = mask_regions(time, flux, regions=[(2458865.5,2458866), (2458990,2458990.5)])
tessplot(time, flux)

#::: 3) estimate variability period
period = estimate_period(time, flux, flux_err)[0]

#::: 4) detrend data
flux = slide_clip(time, flux, high=3, low=20)
flux_flat, trend = flatten(time, flux, method='biweight', window_length=1, return_trend=True)
tessplot(time, flux, trend=trend)
tessplot(time, flux_flat)
period = estimate_period(time, flux_flat, flux_err)[0]

#::: 5) search transits
kwargs = get_tls_kwargs_by_tic(tic_id) #390651552 #459837008
results_all, fig_all = tls_search(time, flux_flat, flux_err, plot=True, 
                                    period_min=1, period_max=20, **kwargs)
for r in results_all:
    print('---------------------')
    pprint(r)
    tessplot(time, flux_flat, trend=r['model'], clip=True)



  """
Pipeline to analyze allesfitter output for planet transit timings

argument 1: allesfitter path
argument 2: p-value threshold
argument 3: Boolean to select to plot wout/with TESS or wout/with/only TESS

Tansu Daylan
MIT Kavli Institute, Cambridge, MA, 02109, US
tansu.daylan@gmail.com
www.tansudaylan.com
"""

import numpy as np
import scipy
import os, datetime, sys

import matplotlib.pyplot as plt
from tdpy.util import summgene 

import allesfitter
import allesfitter.postprocessing.plot_viol
from allesfitter import config

import astropy

class gdatstrt(object):

    def __init__(self):
        self.boollockmodi = False
        pass
    
    def __setattr__(self, attr, valu):
        super(gdatstrt, self).__setattr__(attr, valu)


def plot(gdat, indxstar, indxpara=None, strgtype='evol'):
    
    if indxstar.size == 1:
        strg = gdat.liststrgstar[indxstar[0]] + '_'
    else:
        strg = ''
    
    print('strgtype')
    print(strgtype)
    
    listticklabl = []
    if strgtype == 'epocevol':
        chanlist = [[[] for m in gdat.indxstar] for i in gdat.indxruns]
        xpos = np.array(gdat.listyear)
        for i in gdat.indxruns:
            for m in indxstar:
                chanlist[i][m] = [gdat.timejwst[k][i][m] for k in gdat.indxyear]
        for k in gdat.indxyear:
            listticklabl.append('%s' % str(gdat.listyear[k]))
    else:
        chanlist = []
        numbxdat = gdat.numbruns * indxstar.size
        xpos = 0.6 * (np.arange(numbxdat) + 1.)

        for i in gdat.indxruns:
            for m in indxstar:
                if strgtype == 'jwstcomp':
                    chanlist.append(gdat.timejwst[1][i][m])
                if strgtype == 'paracomp':
                    for k in indxpara:
                        chanlist.append((gdat.listobjtalle[i][m].posterior_params[gdat.liststrgparaconc[k]] - \
                            np.mean(gdat.listobjtalle[i][m].posterior_params[gdat.liststrgparaconc[k]])) * 24. * 60.)
                    
                if strgtype == 'paracomp' or strgtype == 'jwstcomp':
                    ticklabl = '%s, %s' % (gdat.liststrgstar[m], gdat.liststrgruns[i])
                    listticklabl.append(ticklabl)
                else:
                    ticklabl = '%s, %s' % (gdat.liststrgstar[m], gdat.liststrgruns[i])
                    listticklabl.append(ticklabl)
                
        if xpos.size != len(listticklabl):
            raise Exception('')
        
    print('xpos')
    summgene(xpos)
    print('chanlist')
    print(chanlist)
    figr, axis = plt.subplots(figsize=(5, 4))
    if strgtype != 'epocevol':
        axis.violinplot(chanlist, xpos, showmedians=True, showextrema=False)
    else:
        for i in gdat.indxruns:
            for m in indxstar:
                axis.violinplot(chanlist[i][m], xpos, showmedians=True, showextrema=False)

    axis.set_xticks(xpos)
    if strgtype == 'jwstcomp':
        axis.set_ylabel('Transit time residual in 2023 [min]')
        strgbase = strgtype

    if strgtype == 'paracomp':
        if gdat.liststrgparaconc[indxpara] == 'b_period':
            axis.set_ylabel('P [min]')
        else:
            labl = gdat.listlablparaconc[indxpara[0]]
            axis.set_ylabel(labl)
        strgbase = '%04d' % indxpara
    
    if strgtype == 'epocevol':
        axis.set_xlabel('Year')
        axis.set_ylabel('Transit time residual [min]')
        strgbase = strgtype
    
    path = gdat.pathimag + 'viol_%s.%s' % (strgbase, gdat.strgplotextn)
    axis.set_xticklabels(listticklabl)
    plt.tight_layout()
    print('Writing to %s...' % path)
    print()
    figr.savefig(path)
    plt.close()
    

def plot_viol(pathbase, liststrgstar, liststrgruns, lablstrgruns, pathimag, pvalthrs=1e-3):

    strgtimestmp = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    print('allesfitter postprocessing violin plot started at %s...' % strgtimestmp)
    
    # construct global object
    gdat = gdatstrt()
    
    # copy unnamed inputs to the global object
    #for attr, valu in locals().iter():
    for attr, valu in locals().items():
        if '__' not in attr and attr != 'gdat':
            setattr(gdat, attr, valu)

    # runs to be compared for each star
    gdat.numbruns = len(liststrgruns)
    gdat.indxruns = np.arange(gdat.numbruns)
    
    gdat.pathimag = pathimag
    gdat.liststrgstar = liststrgstar

    # stars
    numbstar = len(liststrgstar)
    gdat.indxstar = np.arange(numbstar)

    # plotting
    gdat.strgplotextn = 'png'

    # read parameter keys, labels and posterior from allesfitter output
    liststrgpara = [[] for i in gdat.indxruns]
    listlablpara = [[] for i in gdat.indxruns]
    gdat.listobjtalle = [[[] for m in gdat.indxstar] for i in gdat.indxruns]
    for i in gdat.indxruns:
        for m in gdat.indxstar:
            pathalle = pathbase + '%s/allesfits/allesfit_%s/' % (gdat.liststrgstar[m], gdat.liststrgruns[i])
            print('Reading from %s...' % pathalle)
            config.init(pathalle)
            liststrgpara[i] = np.array(config.BASEMENT.fitkeys)
            listlablpara[i] = np.array(config.BASEMENT.fitlabels)
            # read the chain
            print('pathalle')
            print(pathalle)
            gdat.listobjtalle[i][m] = allesfitter.allesclass(pathalle)
    
    # concatenate the keys, labels from different runs
    gdat.liststrgparaconc = np.concatenate(liststrgpara)
    gdat.liststrgparaconc = np.unique(gdat.liststrgparaconc)
    gdat.listlablparaconc = np.copy(gdat.liststrgparaconc)
    for k, strgparaconc in enumerate(gdat.liststrgparaconc):
        for i, strgruns in enumerate(liststrgruns):
            if strgparaconc in liststrgpara[i]:
                gdat.listlablparaconc[k] = listlablpara[i][np.where(liststrgpara[i] == strgparaconc)[0][0]]
    
    gdat.numbparaconc = len(gdat.liststrgparaconc)
    gdat.indxparaconc = np.arange(gdat.numbparaconc)
    for k, strgpara in enumerate(gdat.liststrgparaconc):
        booltemp = True
        for i in gdat.indxruns:
            if not strgpara in liststrgpara[i]:
                booltemp = False
        if not booltemp:
            continue
        
        ## violin plot
        ## mid-transit time prediction
        plot(gdat, gdat.indxstar, indxpara=np.array([k]), strgtype='paracomp')
        ## per-star 
        #for m in gdat.indxstar:
        #    plot(gdat, indxstar=np.array([m]), indxpara=k, strgtype='paracomp')
        
    # calculate the future evolution of epoch
    gdat.listyear = [2021, 2023, 2025]
    numbyear = len(gdat.listyear)
    gdat.indxyear = np.arange(numbyear)
    gdat.timejwst = [[[[] for m in gdat.indxstar] for i in gdat.indxruns] for k in gdat.indxyear]
    for k, year in enumerate(gdat.listyear):
        epocjwst = astropy.time.Time('%d-01-01 00:00:00' % year, format='iso').jd
        for i in gdat.indxruns:
            for m in gdat.indxstar:
                epoc = gdat.listobjtalle[i][m].posterior_params['b_epoch']
                peri = gdat.listobjtalle[i][m].posterior_params['b_period']
                indxtran = (epocjwst - epoc) / peri
                indxtran = np.mean(np.rint(indxtran))
                if indxtran.size != np.unique(indxtran).size:
                    raise Exception('')
                gdat.timejwst[k][i][m] = epoc + peri * indxtran
                gdat.timejwst[k][i][m] -= np.mean(gdat.timejwst[k][i][m])
                gdat.timejwst[k][i][m] *= 24. * 60.
    
    listfigr = []
    listaxis = []

    # temporal evolution of mid-transit time prediction
    plot(gdat, gdat.indxstar, strgtype='epocevol')
    ## per-star 
    #for m in gdat.indxstar:
    #    plot(gdat, indxstar=np.array([m]), strgtype='epocevol')
    
    ## mid-transit time prediction
    plot(gdat, gdat.indxstar, strgtype='jwstcomp')
    ## per-star 
    #for m in gdat.indxstar:
    #    plot(gdat, indxstar=np.array([m]), strgtype='jwstcomp')
    
    return listfigr, listaxis


    





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 01:10:51 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os, sys
import warnings
from astropy.time import Time
#import pickle
from tqdm import tqdm
warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning) 
warnings.filterwarnings('ignore', category=np.RankWarning) 

#::: allesfitter modules
from . import config
from .utils import latex_printer
from .computer import update_params,\
                     calculate_model, rv_fct, flux_fct,\
                     calculate_baseline,calculate_stellar_var,\
                     calculate_yerr_w,\
                     flux_subfct_sinusoidal_phase_curves
from .exoworlds_rdx.lightcurves import lightcurve_tools as lct
from .exoworlds_rdx.lightcurves.index_transits import get_tmid_observed_transits
                    
                    
 
                     
###############################################################################
#::: print function that prints into console and logfile at the same time
############################################################################### 
def logprint(*text):
    if config.BASEMENT.settings['print_progress']:
        print(*text)
    original = sys.stdout
    try:
        with open( os.path.join(config.BASEMENT.outdir,'logfile_'+config.BASEMENT.now+'.log'), 'a' ) as f:
            sys.stdout = f
            print(*text)
    except OSError:
        pass #For unknown reasons, the combination of open() and os.path.join() does not work on some Windows versions
    sys.stdout = original
                     
    
    
###############################################################################
#::: draw samples from the initial guess
###############################################################################
def draw_initial_guess_samples(Nsamples=1):
    if Nsamples==1:
        samples = np.array([config.BASEMENT.theta_0])
    else:
        samples = config.BASEMENT.theta_0 + config.BASEMENT.init_err * np.random.randn(Nsamples, len(config.BASEMENT.theta_0))    
    return samples
        
    
    
###############################################################################
#::: plot all data in one panel
###############################################################################
def plot_panel(datadir):
    
    config.init(datadir)
    
    if len(config.BASEMENT.settings['inst_phot'])>0 and len(config.BASEMENT.settings['inst_rv'])>0:
        fig, axes = plt.subplots(2,1,figsize=(20,10))
    elif len(config.BASEMENT.settings['inst_phot'])>0:
        fig, axes = plt.subplots(1,1,figsize=(20,5))
        axes = [axes]
    elif len(config.BASEMENT.settings['inst_rv'])>0:
        fig, axes = plt.subplots(1,1,figsize=(20,5))
        axes = [None,axes]
    
    for inst in config.BASEMENT.settings['inst_phot']:
        ax = axes[0]
        ax.plot(config.BASEMENT.fulldata[inst]['time'], config.BASEMENT.fulldata[inst]['flux'], marker='.', ls='none', color='lightgrey', rasterized=True)
        ax.plot(config.BASEMENT.data[inst]['time'], config.BASEMENT.data[inst]['flux'], marker='.', ls='none', label=inst, rasterized=True)
        ax.legend()
        ax.set(ylabel='Relative Flux', xlabel='Time (BJD)')
        
    for inst in config.BASEMENT.settings['inst_rv']:
        ax = axes[1]
        ax.plot(config.BASEMENT.data[inst]['time'], config.BASEMENT.data[inst]['rv'], marker='.', ls='none', label=inst)
        ax.legend()
        ax.set(ylabel='RV (km/s)', xlabel='Time (BJD)')
    
    fig.savefig( os.path.join(config.BASEMENT.outdir,'data_panel.pdf'), bbox_inches='tight' )
    return fig, axes
        



###############################################################################
#::: plot all transits in one panel
###############################################################################
def plot_panel_transits(datadir, ax=None, insts=None, companions=None, colors=None, title=None, ppm=False, ylim=None, yticks=None, fontscale=2):

    config.init(datadir)
    
    #::: more plotting settings
    SMALL_SIZE = 8*fontscale
    MEDIUM_SIZE = 10*fontscale
    BIGGER_SIZE = 12*fontscale
    plt.rc('font', size=MEDIUM_SIZE)          # controls default text sizes
    plt.rc('axes', titlesize=BIGGER_SIZE)     # fontsize of the axes title
    plt.rc('axes', labelsize=BIGGER_SIZE)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize=MEDIUM_SIZE)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=MEDIUM_SIZE)    # fontsize of the tick labels
    plt.rc('legend', fontsize=MEDIUM_SIZE)    # legend fontsize
    plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title
    
    samples = draw_initial_guess_samples()
    params_median, params_ll, params_ul = get_params_from_samples(samples)
    
    if companions is None:
        companions = config.BASEMENT.settings['companions_phot']
    if colors is None:
        colors = [sns.color_palette('deep')[i] for i in [0,1,3]]
    if insts is None:
        insts = config.BASEMENT.settings['inst_phot']

    ally = []
    
    if ax is None:
        ax_init = None
        fig, axes = plt.subplots(len(insts),len(companions),figsize=(6*len(companions),4*len(insts)), sharey=True, sharex=True)
        axes = np.atleast_2d(axes).T
    else:
        ax_init = ax
        axes = np.atleast_2d(ax).T
        
    for i,(companion, color) in enumerate(zip(companions, colors)):
        
        for j,inst in enumerate(insts):
            ax = axes[i,j]
            
            key='flux'
            if title is None:
                if i==0:
                    title=inst
                else:
                    title=''
            if j==len(insts)-1:
                xlabel=r'$\mathrm{ T - T_0 \ (h) }$'
            else:
                xlabel=''
            if i==0:
                if ppm:
                    ylabel=r'$\Delta$ Flux (ppm)'
                else:
                    ylabel='Relative Flux'
            else:
                ylabel=''
            alpha = 1.
                    
            x = config.BASEMENT.data[inst]['time']
            baseline_median = calculate_baseline(params_median, inst, key) #evaluated on x (!)
            y = config.BASEMENT.data[inst][key] - baseline_median
            
            zoomfactor = params_median[companion+'_period']*24.
            
            for other_companion in config.BASEMENT.settings['companions_phot']:
                if companion!=other_companion:
                    model = flux_fct(params_median, inst, other_companion)
                    y -= model
                    y += 1.
            
            if ppm:
                y = (y-1)*1e6
                    
            dt = 20./60./24. / params_median[companion+'_period']
                
            phase_time, phase_y, phase_y_err, _, phi = lct.phase_fold(x, y, params_median[companion+'_period'], params_median[companion+'_epoch'], dt = dt, ferr_type='meansig', ferr_style='sem', sigmaclip=True)    
            ax.plot( phi*zoomfactor, y, 'b.', color='silver', rasterized=True )
            ax.errorbar( phase_time*zoomfactor, phase_y, yerr=phase_y_err, linestyle='none', marker='o', ms=8, color=color, capsize=0, zorder=11 )
            ax.set_xlabel(xlabel, fontsize=BIGGER_SIZE)
            ax.set_ylabel(ylabel, fontsize=BIGGER_SIZE)

            ax.text(0.97,0.87,companion,ha='right',va='bottom',transform=ax.transAxes,fontsize=BIGGER_SIZE)
            ax.text(0.03,0.87,title,ha='left',va='bottom',transform=ax.transAxes,fontsize=MEDIUM_SIZE)
        
            ally += list(phase_y)
            
            #model, phased
            xx = np.linspace( -4./zoomfactor, 4./zoomfactor, 1000)
            xx2 = params_median[companion+'_epoch'] + np.linspace( -4./zoomfactor, 4./zoomfactor, 1000)*params_median[companion+'_period']
            for ii in range(samples.shape[0]):
                s = samples[ii,:]
                p = update_params(s)
#                p = update_params(s, phased=True)
                model = flux_fct(p, inst, companion, xx=xx2) #evaluated on xx2 (!)
                if ppm:
                    model = (model-1)*1e6
                ax.plot( xx*zoomfactor, model, 'r-', alpha=alpha, zorder=12, lw=2 )
                 
    if ppm:
        ylim0 = np.nanmin(ally) - 500
        ylim1 = np.nanmax(ally) + 500
    else:
        ylim0 = np.nanmin(ally) - 500/1e6
        ylim1 = np.nanmax(ally) + 500/1e6
   
    if ylim is None:
        ylim = [ylim0, ylim1]
    
    for i in range(len(companions)):
        for j in range(len(insts)):
            ax = axes[i,j]
            ax.set( xlim=[-4,4], ylim=ylim )
            if yticks is not None:
                ax.set(yticks=yticks)
            ax.set_xticklabels(ax.get_xticks(), {'fontsize': MEDIUM_SIZE})
            ax.set_yticklabels(ax.get_yticks(), {'fontsize': MEDIUM_SIZE})
            
    
    plt.tight_layout()

    if ax_init is None:
        fig.savefig( os.path.join(config.BASEMENT.outdir,'data_panel_transits.pdf'), bbox_inches='tight' )
        return fig, axes
    else:
        return ax




###############################################################################
#::: plot
###############################################################################
def afplot(samples, companion):
    '''
    Inputs:
    -------
    samples : array
        samples from the initial guess, or from the MCMC / Nested Sampling posteriors
    '''
#    global config.BASEMENT
    
    if config.BASEMENT.settings['fit_ttvs'] is False:
      
        N_inst = len(config.BASEMENT.settings['inst_all'])
        
        if 'do_not_phase_fold' in config.BASEMENT.settings and config.BASEMENT.settings['do_not_phase_fold']:
            fig, axes = plt.subplots(N_inst,1,figsize=(6*1,4*N_inst))
            styles = ['full']
        elif config.BASEMENT.settings['phase_curve']:
            fig, axes = plt.subplots(N_inst,5,figsize=(6*5,4*N_inst))
            styles = ['full','phase','phase_curve','phasezoom','phasezoom_occ']
        elif config.BASEMENT.settings['secondary_eclipse']:
            fig, axes = plt.subplots(N_inst,4,figsize=(6*4,4*N_inst))
            styles = ['full','phase','phasezoom','phasezoom_occ']
        else:
            fig, axes = plt.subplots(N_inst,3,figsize=(6*3,4*N_inst))
            styles = ['full','phase','phasezoom']
        axes = np.atleast_2d(axes)
        
        for i,inst in enumerate(config.BASEMENT.settings['inst_all']):
            for j,style in enumerate(styles):
    #            print(i,j,inst,style)
                #::: don't phase-fold single day photometric follow-up
    #            if (style=='phase') & (inst in config.BASEMENT.settings['inst_phot']) & ((config.BASEMENT.data[inst]['time'][-1] - config.BASEMENT.data[inst]['time'][0]) < 1.):
    #                axes[i,j].axis('off')
                #::: don't zoom onto RV data
                if ('zoom' in style) & (inst in config.BASEMENT.settings['inst_rv']):
                    axes[i,j].axis('off')
                #::: don't plot if the companion is not covered by an instrument
                elif (inst in config.BASEMENT.settings['inst_phot']) & (companion not in config.BASEMENT.settings['companions_phot']):
                    axes[i,j].axis('off')
                #::: don't plot if the companion is not covered by an instrument
                elif (inst in config.BASEMENT.settings['inst_rv']) & (companion not in config.BASEMENT.settings['companions_rv']):
                    axes[i,j].axis('off')
                else:
                    plot_1(axes[i,j], samples, inst, companion, style)
    
        plt.tight_layout()
        return fig, axes

    else:
        return None, None
    


###############################################################################
#::: plot_1 (helper function)
###############################################################################
def plot_1(ax, samples, inst, companion, style, 
           base=None, dt=None,
           zoomwindow=8., force_binning=False,
           kwargs_data=None,
           kwargs_model=None,
           kwargs_ax=None):
    '''
    Inputs:
    -------
    ax : matplotlib axis
    
    samples : array
        Prior or posterior samples to plot the fit from
    
    inst: str
        Name of the instrument (e.g. 'TESS')
        
    companion : None or str
        None or 'b'/'c'/etc.
        
    style: str
        'full' / 'per_transit' / 'phase' / 'phasezoom' / 'phasezoom_occ' /'phase_curve'
        'full_residuals' / 'phase_residuals' / 'phasezoom_residuals' / 'phasezoom_occ_residuals' / 'phase_curve_residuals'
    
    zoomwindow: int or float
        the full width of the window to zoom into (in hours)
        default: 8 hours
    
    base: a BASEMENT class object
        (for internal use only)
        
    dt : float
        time steps on which the model should be evaluated for plots
        in days
        default for style='full': 2 min for <1 day of data; 30 min for >1 day of data.
        
    Notes:
    ------
    yerr / epoch / period: 
        come either from
        a) the initial_guess value or 
        b) the MCMC median,
        depending on what is plotted (i.e. not from individual samples)

    '''

    #==========================================================================
    #::: interpret input
    #==========================================================================
    if base==None:
        base = config.BASEMENT
    
    if samples is not None:
        params_median, params_ll, params_ul = get_params_from_samples(samples)
    
    if kwargs_data is None: kwargs_data = {}
    if 'label' not in kwargs_data: kwargs_data['label'] = inst
    if 'marker' not in kwargs_data: kwargs_data['marker'] = '.'
    if 'markersize' not in kwargs_data: kwargs_data['markersize'] = 8.
    if 'linestyle' not in kwargs_data: kwargs_data['linestyle'] = 'none'
    if 'color' not in kwargs_data: kwargs_data['color'] = 'b'
    if 'alpha' not in kwargs_data: kwargs_data['alpha'] = 1.
    if 'rasterized' not in kwargs_data: kwargs_data['rasterized'] = True
    
    if kwargs_model is None: kwargs_model = {}
    if 'marker' not in kwargs_model: kwargs_model['marker'] = 'none'
    if 'markersize' not in kwargs_model: kwargs_model['markersize'] = 0.
    if 'linestyle' not in kwargs_model: kwargs_model['linestyle'] = '-'
    if 'color' not in kwargs_model: kwargs_model['color'] = 'r'
    if 'alpha' not in kwargs_model: kwargs_model['alpha'] = 1.
    
    if kwargs_ax is None: kwargs_ax = {}
    if 'title' not in kwargs_ax: kwargs_ax['title'] = None
    if 'xlabel' not in kwargs_ax: kwargs_ax['xlabel'] = None
    if 'ylabel' not in kwargs_ax: kwargs_ax['ylabel'] = None
    
    timelabel = 'Time' #removed feature
    
    
    #==========================================================================
    #::: helper fct
    #==========================================================================
    def set_title(title1):
        if kwargs_ax['title'] is None: return title1
        else: return kwargs_ax['title']
    
    
    #==========================================================================
    #::: do stuff
    #==========================================================================
    if inst in base.settings['inst_phot']:
        key='flux'
        baseline_plus = 1.
        if style in ['full']:
            ylabel = 'Relative Flux'
        elif style in ['full_minus_offset']:
            ylabel = 'Relative Flux - Offset'
        elif style in ['phase', 'phasezoom', 'phasezoom_occ', 'phase_curve']:
            ylabel = 'Relative Flux - Baseline'
        elif style in ['full_residuals', 'phase_residuals', 'phasezoom_residuals', 'phasezoom_occ_residuals', 'phase_curve_residuals']:
            ylabel = 'Residuals'
    elif inst in base.settings['inst_rv']:
        key='rv'
        baseline_plus = 0.
        if style in ['full']:
            ylabel = 'RV (km/s)'
        elif style in ['full_minus_offset']:
            ylabel = 'RV (km/s) - Offset'
        elif style in ['phase', 'phasezoom', 'phasezoom_occ', 'phase_curve']:
            ylabel = 'RV (km/s) - Baseline'
        elif style in ['full_residuals', 'phase_residuals', 'phasezoom_residuals', 'phasezoom_occ_residuals', 'phase_curve_residuals']:
            ylabel = 'Residuals'
        
    else:
        raise ValueError('inst should be listed in inst_phot or inst_rv...')
    
    
    if samples is not None:
        if samples.shape[0]==1:
            alpha = 1.
        else:
            alpha = 0.1
        
    
    #==========================================================================
    # full time series, not phased
    # plot the 'undetrended' data
    # plot each sampled model + its baseline 
    #==========================================================================
    if style in ['full', 'full_minus_offset', 'full_residuals']:
        
        #::: set it up
        x = base.data[inst]['time']
        
        if timelabel=='Time_since':
            x = np.copy(x)
            objttime = Time(x, format='jd', scale='utc')
            xsave = np.copy(x)
            x -= x[0]

        y = 1.*base.data[inst][key]
        yerr_w = calculate_yerr_w(params_median, inst, key)
        
        
        #::: remove offset only (if wished)
        if style in ['full_minus_offset']:
            baseline = calculate_baseline(params_median, inst, key)
            y -= np.median(baseline)
            
            
        #::: calculate residuals (if wished)
        if style in ['full_residuals']:
            model = calculate_model(params_median, inst, key)
            baseline = calculate_baseline(params_median, inst, key)
            stellar_var = calculate_stellar_var(params_median, 'all', key, xx=x)
            y -= model+baseline+stellar_var
            
            
        #::: plot data, not phase        
#        ax.errorbar(base.fulldata[inst]['time'], base.fulldata[inst][key], yerr=np.nanmedian(yerr_w), marker='.', linestyle='none', color='lightgrey', zorder=-1, rasterized=True ) 
        # ax.errorbar(x, y, yerr=yerr_w, marker=kwargs_data['marker'], markersize=kwargs_data['markersize'], linestyle=kwargs_data['linestyle'], color=kwargs_data['color'], alpha=kwargs_data['alpha'], capsize=0, rasterized=kwargs_data['rasterized'] )  
        ax.errorbar(x, y, yerr=yerr_w, capsize=0, **kwargs_data)  
        if base.settings['color_plot']:
            ax.scatter(x, y, c=x, marker='o', rasterized=kwargs_data['rasterized'], cmap='inferno', zorder=11 ) 
            
        if timelabel=='Time_since':
            ax.set(xlabel='Time since %s [days]' % objttime[0].isot[:10], ylabel=ylabel, title=set_title(inst))
        elif timelabel=='Time':
            ax.set(xlabel='Time (BJD)', ylabel=ylabel, title=set_title(inst))
            
            
        #::: plot model + baseline, not phased
        if (style in ['full','full_minus_offset']) and (samples is not None):
            
            #if <1 day of photometric data: plot with 2 min resolution
            if dt is None:
                if ((x[-1] - x[0]) < 1): 
                    dt = 2./24./60. 
                #else: plot with 30 min resolution
                else: 
                    dt = 30./24./60. 
                    
            if key == 'flux':
                xx_full = np.arange( x[0], x[-1]+dt, dt)
                Npoints_chunk = 48
                for i_chunk in tqdm(range(int(1.*len(xx_full)/Npoints_chunk)+2)):
                    xx = xx_full[i_chunk*Npoints_chunk:(i_chunk+1)*Npoints_chunk] #plot in chunks of 48 points (1 day)
                    if len(xx)>0 and any( (x>xx[0]) & (x<xx[-1]) ): #plot only where there is data
                        for i in range(samples.shape[0]):
                            s = samples[i,:]
                            p = update_params(s)
                            model = calculate_model(p, inst, key, xx=xx) #evaluated on xx (!)
                            baseline = calculate_baseline(p, inst, key, xx=xx) #evaluated on xx (!)
                            if style in ['full_minus_offset']:
                                baseline -= np.median(baseline)
                            stellar_var = calculate_stellar_var(p, 'all', key, xx=xx) #evaluated on xx (!)
                            ax.plot( xx, baseline+stellar_var+baseline_plus, 'k-', color='orange', alpha=alpha, zorder=12 )
                            ax.plot( xx, model+baseline+stellar_var, 'r-', alpha=alpha, zorder=12 )
            elif key=='rv':
                xx = np.arange( x[0], x[-1]+dt, dt)
                for i in range(samples.shape[0]):
                    s = samples[i,:]
                    p = update_params(s)
                    model = calculate_model(p, inst, key, xx=xx) #evaluated on xx (!)
                    baseline = calculate_baseline(p, inst, key, xx=xx) #evaluated on xx (!)
                    if style in ['full_minus_offset']:
                        baseline -= np.median(baseline)                    
                    stellar_var = calculate_stellar_var(p, 'all', key, xx=xx) #evaluated on xx (!)
                    ax.plot( xx, baseline+stellar_var+baseline_plus, 'k-', color='orange', alpha=alpha, zorder=12 )
                    ax.plot( xx, model+baseline+stellar_var, 'r-', alpha=alpha, zorder=12 )
        
        #::: other stuff
        if timelabel=='Time_since':
            x = np.copy(xsave)
            
            
            
            
    #==========================================================================
    # phase-folded time series
    # get a 'median' baseline from intial guess value / MCMC median result
    # detrend the data with this 'median' baseline
    # then phase-fold the 'detrended' data
    # plot each phase-folded model (without baseline)
    # Note: this is not ideal, as we overplot models with different epochs/periods/baselines onto a phase-folded plot
    #==========================================================================
    elif style in ['phase', 'phasezoom', 'phasezoom_occ', 'phase_curve',
                   'phase_residuals', 'phasezoom_residuals', 'phasezoom_occ_residuals', 'phase_curve_residuals']:
        
        #::: data - baseline_median
        x = 1.*base.data[inst]['time']
        baseline_median = calculate_baseline(params_median, inst, key) #evaluated on x (!)
        stellar_var_median = calculate_stellar_var(params_median, 'all', key, xx=x) #evaluated on x (!)
        y = base.data[inst][key] - baseline_median - stellar_var_median
        yerr_w = calculate_yerr_w(params_median, inst, key)
        
        #::: zoom?
        if style in ['phasezoom', 'phasezoom_occ', 
                     'phasezoom_residuals', 'phasezoom_occ_residuals']: 
            zoomfactor = params_median[companion+'_period']*24.
        else: 
            zoomfactor = 1.
        
        
        #----------------------------------------------------------------------
        #::: Radial velocity
        #::: need to take care of multiple companions
        #----------------------------------------------------------------------
        if (inst in base.settings['inst_rv']):
            
            #::: remove other companions
            for other_companion in base.settings['companions_rv']:
                if companion!=other_companion:
                    model = rv_fct(params_median, inst, other_companion)[0]
                    y -= model
            
            
            #::: calculate residuals (if wished)
            if style in ['phase_residuals', 'phasezoom_residuals', 'phasezoom_occ_residuals', 'phase_curve_residuals']:
                model = rv_fct(params_median, inst, companion)[0]
                y -= model
                
                
            #::: plot data, phased        
            phase_time, phase_y, phase_y_err, _, phi = lct.phase_fold(x, y, params_median[companion+'_period'], params_median[companion+'_epoch'], dt = 0.002, ferr_type='meansig', ferr_style='sem', sigmaclip=False)    
            if (len(x) > 500) or force_binning:
                ax.plot( phi*zoomfactor, y, 'k.', color='lightgrey', rasterized=kwargs_data['rasterized'] ) #don't allow any other kwargs_data here
                ax.errorbar( phase_time*zoomfactor, phase_y, yerr=phase_y_err, capsize=0, zorder=11, **kwargs_data )
            else:
                ax.errorbar( phi*zoomfactor, y, yerr=yerr_w, capsize=0, zorder=11, **kwargs_data )      
            ax.set(xlabel='Phase', ylabel=ylabel, title=set_title(inst+', companion '+companion+' only'))
    
    
            #::: plot model, phased (if wished)
            if (style in ['phase', 'phasezoom', 'phasezoom_occ', 'phase_curve']) and (samples is not None):
                xx = np.linspace( -0.25, 0.75, 1000)
                xx2 = params_median[companion+'_epoch']+np.linspace( -0.25, 0.75, 1000)*params_median[companion+'_period']
                for i in range(samples.shape[0]):
                    s = samples[i,:]
                    p = update_params(s)
#                    p = update_params(s, phased=True)
                    model = rv_fct(p, inst, companion, xx=xx2)[0]
                    ax.plot( xx*zoomfactor, model, 'r-', alpha=alpha, zorder=12 )
            
        
        #----------------------------------------------------------------------
        #::: Photometry
        #----------------------------------------------------------------------
        elif (inst in base.settings['inst_phot']):
            
            #::: remove other companions
            for other_companion in base.settings['companions_phot']:
                if companion!=other_companion:
                    model = flux_fct(params_median, inst, other_companion)
                    y -= (model-1.)
                    
                    
            #::: calculate residuals (if wished)
            if style in ['phase_residuals', 'phasezoom_residuals', 'phasezoom_occ_residuals', 'phase_curve_residuals']:
                model = flux_fct(params_median, inst, companion)
                y -= model
                    
                
            #::: plot data, phased  
            if style in ['phase', 
                         'phase_residuals']:
                dt = 0.002
            elif style in ['phase_curve', 
                           'phase_curve_residuals']:
                dt = 0.01            
            elif style in ['phasezoom', 'phasezoom_occ', 
                           'phasezoom_residuals', 'phasezoom_occ_residuals']: 
                dt = 15./60./24. / params_median[companion+'_period']
                
            phase_time, phase_y, phase_y_err, _, phi = lct.phase_fold(x, y, params_median[companion+'_period'], params_median[companion+'_epoch'], dt = dt, ferr_type='meansig', ferr_style='sem', sigmaclip=False)    
            if (len(x) > 500) or force_binning:
                if style in ['phase_curve', 
                             'phase_curve_residuals']:
                    ax.plot( phase_time*zoomfactor, phase_y, 'b.', color=kwargs_data['color'], rasterized=kwargs_data['rasterized'], zorder=11 )                    
                else: 
                    ax.plot( phi*zoomfactor, y, 'b.', color='lightgrey', rasterized=kwargs_data['rasterized'], )
                    ax.errorbar( phase_time*zoomfactor, phase_y, yerr=phase_y_err, capsize=0, zorder=11, **kwargs_data )
            else:
                ax.errorbar( phi*zoomfactor, y, yerr=yerr_w, capsize=0, zorder=11, **kwargs_data )
                if base.settings['color_plot']:
                    ax.scatter( phi*zoomfactor, y, c=x, marker='o', rasterized=kwargs_data['rasterized'], cmap='inferno', zorder=11 )          
            ax.set(xlabel='Phase', ylabel=ylabel, title=set_title(inst+', companion '+companion))
    
    
            #::: plot model, phased (if wished)
            if style in ['phase', 'phasezoom', 'phasezoom_occ', 'phase_curve']:
                
                if style in ['phase', 'phase_curve']:
                    xx = np.linspace(-0.25, 0.75, 1000)
                    xx2 = params_median[companion+'_epoch'] + xx * params_median[companion+'_period']
                elif style in ['phasezoom']:
                    xx = np.linspace( -10./zoomfactor, 10./zoomfactor, 1000)
                    xx2 = params_median[companion+'_epoch'] + xx * params_median[companion+'_period']
                elif style in ['phasezoom_occ']:
                    e = params_median[companion+'_f_s']**2 + params_median[companion+'_f_c']**2
                    w = np.mod( np.arctan2(params_median[companion+'_f_s'], params_median[companion+'_f_c']), 2*np.pi) #in rad, from 0 to 2*pi
                    phase_shift = 0.5 * (1. + 4./np.pi * e * np.cos(w)) #in phase units; approximation from Winn2010
                    xx = np.linspace( -10./zoomfactor + phase_shift, 10./zoomfactor + phase_shift, 1000 )
                    xx2 = params_median[companion+'_epoch'] + xx * params_median[companion+'_period']
    
                if samples is not None:
                    for i in range(samples.shape[0]):
                        s = samples[i,:]
                        p = update_params(s)
    #                    p = update_params(s, phased=True)
                        model = flux_fct(p, inst, companion, xx=xx2) #evaluated on xx (!)
                        ax.plot( xx*zoomfactor, model, 'r-', alpha=alpha, zorder=12 )
             
        
        #::: x-zoom?
        if style in ['phasezoom',
                     'phasezoom_residuals']:
                ax.set( xlim=[-zoomwindow/2.,zoomwindow/2.], xlabel=r'$\mathrm{ T - T_0 \ (h) }$' )
        elif style in ['phasezoom_occ',
                       'phasezoom_occ_residuals']:
                xlower = -zoomwindow/2. + phase_shift*params_median[companion+'_period']*24.
                xupper = zoomwindow/2. + phase_shift*params_median[companion+'_period']*24.
                ax.set( xlim=[xlower, xupper], xlabel=r'$\mathrm{ T - T_0 \ (h) }$' )
        
        
        #::: y-zoom onto occultation and phase variations
        if style in ['phasezoom_occ']:
                # try:
            buf = phase_y[phase_time>0.25] #TODO: replace with proper eclipse indexing
            def nanptp(arr): return np.nanmax(arr)-np.nanmin(arr)
            ax.set( ylim=[np.nanmin(buf)-0.1*nanptp(buf), np.nanmax(buf)+0.1*nanptp(buf)] )
                # except:
                #     ax.set( ylim=[0.999,1.0005] )
       
        if style in ['phase_curve', 
                     'phase_curve_residuals']:
                try:
                    phase_curve_no_dips = flux_subfct_sinusoidal_phase_curves(params_median, inst, companion, np.ones_like(xx), xx=xx)
                    ax.set(ylim=[np.min(phase_curve_no_dips)-0.1*np.ptp(phase_curve_no_dips), np.max(phase_curve_no_dips)+0.1*np.ptp(phase_curve_no_dips)])
                except:
                    ax.set( ylim=[0.999,1.001] )




    
###############################################################################
#::: plot individual transits
###############################################################################
def afplot_per_transit(samples, inst, companion, base=None, kwargs_dict=None):
        
    #==========================================================================
    #::: input
    #==========================================================================
    if base==None: base = config.BASEMENT
    if kwargs_dict is None: kwargs_dict = {}
    if 'window' not in kwargs_dict: kwargs_dict['window'] = 8./24. # in days
    if 'rasterized' not in kwargs_dict: kwargs_dict['rasterized'] = True
    if 'marker' not in kwargs_dict: kwargs_dict['marker'] = '.'
    if 'linestyle' not in kwargs_dict: kwargs_dict['linestyle'] = 'none'
    if 'color' not in kwargs_dict: kwargs_dict['color'] = 'b'
    if 'markersize' not in kwargs_dict: kwargs_dict['markersize'] = 8
    
    
    #==========================================================================
    #::: translate input
    #==========================================================================
    window = kwargs_dict['window']
    rasterized = kwargs_dict['rasterized']
    marker = kwargs_dict['marker']
    linestyle = kwargs_dict['linestyle']
    color = kwargs_dict['color']
    markersize = kwargs_dict['markersize']
    
        
    #==========================================================================
    #::: configurations
    #==========================================================================
    if inst in base.settings['inst_phot']:
        key = 'flux'
        ylabel = 'Realtive Flux'
        baseline_plus = 1.
    elif inst in base.settings['inst_rv']:
        key = 'rv'   
        ylabel = 'RV (km/s)'
        
    if samples.shape[0]==1:
        alpha = 1.
    else:
        alpha = 0.1
        
        
    #==========================================================================
    #::: load data and models
    #==========================================================================
    params_median, params_ll, params_ul = get_params_from_samples(samples)
    x = base.data[inst]['time']
    y = 1.*base.data[inst][key]
    yerr_w = calculate_yerr_w(params_median, inst, key)
    
    tmid_observed_transits = get_tmid_observed_transits(x, params_median[companion+'_epoch'], params_median[companion+'_period'], base.settings['fast_fit_width'])
    N_transits = len(tmid_observed_transits)
    
    fig, axes = plt.subplots(N_transits, 1, figsize=(6,4*N_transits), sharey=True, tight_layout=True)
    
    if N_transits>0:
        axes = np.atleast_1d(axes)
        axes[0].set(title=inst)
        
        for i, t in enumerate(tmid_observed_transits):
            ax = axes[i]
            
            #::: mark data
            ind = np.where((x >= (t - window/2.)) & (x <= (t + window/2.)))[0]
            
            #::: plot model
            ax.errorbar(x[ind], y[ind], yerr=yerr_w[ind], marker=marker, linestyle=linestyle, color=color, markersize=markersize, alpha=1, capsize=0, rasterized=rasterized)  

            #::: plot model + baseline, not phased
            dt = 2./24./60. 
            xx = np.arange(x[ind][0], x[ind][-1]+dt, dt)
            for j in range(samples.shape[0]):
                s = samples[j,:]
                p = update_params(s)
                model = calculate_model(p, inst, key, xx=xx) #evaluated on xx (!)
                baseline = calculate_baseline(p, inst, key, xx=xx) #evaluated on xx (!)
                stellar_var = calculate_stellar_var(p, 'all', key, xx=xx) #evaluated on xx (!)
                ax.plot( xx, baseline+stellar_var+baseline_plus, 'k-', color='orange', alpha=alpha, zorder=12 )
                ax.plot( xx, model+baseline+stellar_var, 'r-', alpha=alpha, zorder=12 )
            ax.set(xlim=[t-window/2., t+window/2.])
            ax.axvline(t,color='grey',lw=2,ls='--',label='linear prediction')
            if base.settings['fit_ttvs']==True:
                ax.axvline(t+params_median[companion+'_ttv_transit_'+str(i+1)],color='r',lw=2,ls='--',label='TTV midtime')
            
            #::: axes decoration
            ax.set(xlabel='Time', ylabel=ylabel)
            ax.text(0.95,0.95,'Transit '+str(i+1), va='top', ha='right', transform=ax.transAxes)
            
        if base.settings['fit_ttvs']==True:
            axes[0].legend(loc='upper left')
            
    else:
        warnings.warn('No transit of companion '+companion+' for '+inst+'.')
    
    return fig, axes
            
                


    
###############################################################################
#::: update params with MCMC/NS results
###############################################################################
def get_params_from_samples(samples):
    '''
    read MCMC results and update params
    '''
    theta_median = np.percentile(samples, 50, axis=0)
    theta_ul = np.percentile(samples, 84, axis=0) - theta_median
    theta_ll = theta_median - np.percentile(samples, 16, axis=0)
    params_median = update_params(theta_median)
    params_ll = update_params(theta_ll)
    params_ul = update_params(theta_ul)
    
    return params_median, params_ll, params_ul



###############################################################################
#::: save table
###############################################################################
def save_table(samples, mode):
    '''
    Inputs:
    -------
    samples : array
        posterior samples
    mode : string
        'mcmc' or 'ns'
    '''
    
    params, params_ll, params_ul = get_params_from_samples(samples)
    
    with open( os.path.join(config.BASEMENT.outdir,mode+'_table.csv'), 'w' ) as f:
        f.write('#name,median,lower_error,upper_error,label,unit\n')
        f.write('#Fitted parameters,,,\n')
        for i, key in enumerate(config.BASEMENT.allkeys):
            if key not in config.BASEMENT.fitkeys:
                f.write(key + ',' + str(params[key]) + ',' + '(fixed),(fixed),'+config.BASEMENT.labels[i]+','+config.BASEMENT.units[i]+'\n')
            else:
                f.write(key + ',' + str(params[key]) + ',' + str(params_ll[key]) + ',' + str(params_ul[key]) + ',' + config.BASEMENT.labels[i] + ',' + config.BASEMENT.units[i] + '\n' )
   
        
        
###############################################################################
#::: save Latex table
###############################################################################
def save_latex_table(samples, mode):
    '''
    Inputs:
    -------
    samples : array
        posterior samples
    mode : string
        'mcmc' or 'ns'
    '''
    
    params_median, params_ll, params_ul = get_params_from_samples(samples)
        
    with open(os.path.join(config.BASEMENT.outdir,mode+'_latex_table.txt'),'w') as f,\
         open(os.path.join(config.BASEMENT.outdir,mode+'_latex_cmd.txt'),'w') as f_cmd:
            
        f.write('parameter & value & unit & fit/fixed \\\\ \n')
        f.write('\\hline \n')
        f.write('\\multicolumn{4}{c}{\\textit{Fitted parameters}} \\\\ \n')
        f.write('\\hline \n')
        
        for i, key in enumerate(config.BASEMENT.allkeys):
            if key not in config.BASEMENT.fitkeys:                
                value = str(params_median[key])
                f.write(config.BASEMENT.labels[i] + ' & $' + value + '$ & '  + config.BASEMENT.units[i] + '& fixed \\\\ \n')            
                simplename = key.replace("_", "").replace("/", "over").replace("(", "").replace(")", "").replace("1", "one").replace("2", "two").replace("3", "three")
                comment = config.BASEMENT.labels[i] + '$=' + value + '$ ' + config.BASEMENT.units[i]
                comment = comment.replace("$$","")
                f_cmd.write('\\newcommand{\\'+simplename+'}{$'+value+'$} %'+comment+'\n')
                # f_cmd.write('\\newcommand{\\'+key.replace("_", "")+'}{$'+value+'$} %'+label+' = '+value+'\n')

            else:            
                value = latex_printer.round_tex(params_median[key], params_ll[key], params_ul[key])
                f.write(config.BASEMENT.labels[i] + ' & $' + value + '$ & ' + config.BASEMENT.units[i] + '& fit \\\\ \n' )
                simplename = key.replace("_", "").replace("/", "over").replace("(", "").replace(")", "").replace("1", "one").replace("2", "two").replace("3", "three")
                comment = config.BASEMENT.labels[i] + '$=' + value + '$ ' + config.BASEMENT.units[i]
                comment = comment.replace("$$","")
                f_cmd.write('\\newcommand{\\'+simplename+'}{$'+value+'$} %'+comment+'\n')
                 # f_cmd.write('\\newcommand{\\'+key.replace("_", "")+'}{$='+value+'$} %'+label+' = '+value+'\n')


    
###############################################################################
#::: show initial guess
###############################################################################
def show_initial_guess(datadir, quiet=False, do_logprint=True, do_plot=True, return_figs=False, kwargs_dict=None):
    #::: init
    config.init(datadir, quiet=quiet)    
    
    #::: show initial guess
    if do_logprint: 
        logprint_initial_guess()
    if do_plot: 
        return plot_initial_guess(return_figs=return_figs, kwargs_dict=kwargs_dict)
    
    

###############################################################################
#::: logprint initial guess
###############################################################################
def logprint_initial_guess():
    '''
    Inputs:
    -------
    datadir : str
        the working directory for allesfitter
        must contain all the data files
        output directories and files will also be created inside datadir
            
    Outputs:
    --------
    This will output information into the console, 
    and create a file called datadir/results/initial_guess.pdf
    '''
        
    logprint('\nSettings:')
    logprint('--------------------------')
    for key in config.BASEMENT.settings:
        if config.BASEMENT.settings[key]!='':
            logprint('{0: <30}'.format(key), '{0: <15}'.format(str(config.BASEMENT.settings[key]))) #I hate Python 3, and I hope someone finds this comment...
        else:
            logprint('\n{0: <30}'.format(key))

    logprint('\nParameters:')
    logprint('--------------------------')    
    for i, key in enumerate(config.BASEMENT.params):
        if key in config.BASEMENT.fitkeys: 
            ind = np.where( config.BASEMENT.fitkeys == key )[0][0]
            logprint('{0: <30}'.format(key), '{0: <15}'.format(str(config.BASEMENT.params[key])), '{0: <5}'.format('free'), '{0: <30}'.format(str(config.BASEMENT.bounds[ind])) )
        else: 
            if config.BASEMENT.params[key]!='':
                logprint('{0: <30}'.format(key), '{0: <15}'.format(str(config.BASEMENT.params[key])), '{0: <5}'.format('set'))
            else:
                logprint('\n{0: <30}'.format(key))
    
    logprint('\nExternal priors:')
    logprint('--------------------------')  
    if 'host_density' in config.BASEMENT.external_priors:
        logprint('\nStellar density prior (automatically set):', config.BASEMENT.external_priors['host_density'], '(g cm^-3)')
    else:
        logprint('No external priors defined.')
    
    logprint('\nndim:', config.BASEMENT.ndim)



###############################################################################
#::: plot initial guess
###############################################################################
def plot_initial_guess(return_figs=False, kwargs_dict=None):
    
    samples = draw_initial_guess_samples()
    
    if return_figs==False:
        for companion in config.BASEMENT.settings['companions_all']:
            fig, axes = afplot(samples, companion)
            if fig is not None:
                fig.savefig( os.path.join(config.BASEMENT.outdir,'initial_guess_'+companion+'.pdf'), bbox_inches='tight' )
                plt.close(fig)
            
        for companion in config.BASEMENT.settings['companions_phot']:
            for inst in config.BASEMENT.settings['inst_phot']:
                try:
                    fig, axes = afplot_per_transit(samples, inst, companion, kwargs_dict=kwargs_dict)
                    fig.savefig( os.path.join(config.BASEMENT.outdir,'initial_guess_per_transit_'+inst+'_'+companion+'.pdf'), bbox_inches='tight' )
                    plt.close(fig)
                except:
                    pass
        return None
    
    else:
        fig_list = []
        for companion in config.BASEMENT.settings['companions_all']:
            fig, axes = afplot(samples, companion)
            fig_list.append(fig)
        return fig_list
            
    
    
    
###############################################################################
#::: plot initial guess
###############################################################################
def plot_ttv_results(params_median, params_ll, params_ul):
    for companion in config.BASEMENT.settings['companions_all']:
        fig, axes = plt.subplots()
        axes.axhline(0, color='grey', linestyle='--')
        for i in range(len(config.BASEMENT.data[companion+'_tmid_observed_transits'])):
            axes.errorbar( i+1, params_median[companion+'_ttv_transit_'+str(i+1)]*24*60, 
                           yerr=np.array([[ params_ll[companion+'_ttv_transit_'+str(i+1)]*24*60, params_ul[companion+'_ttv_transit_'+str(i+1)]*24*60 ]]).T, 
                           color=config.BASEMENT.settings[companion+'_color'], fmt='.')
        axes.set(xlabel='Tranist Nr.', ylabel='TTV (mins)')
        fig.savefig( os.path.join(config.BASEMENT.outdir,'ttv_results_'+companion+'.pdf'), bbox_inches='tight' )
        plt.close(fig)
    
    
    
    
###############################################################################
#::: get latex labels
###############################################################################
def get_labels(datadir, as_type='dic'):
    config.init(datadir)
    
    if as_type=='2d_array':
        return config.BASEMENT.labels
        
    if as_type=='dic':
        labels_dic = {}
        for key in config.BASEMENT.fitkeys:
            ind = np.where(config.BASEMENT.allkeys==key)[0]
            labels_dic[key] = config.BASEMENT.labels[ind][0]
        return labels_dic
    


###############################################################################
#::: top-level user interface
###############################################################################
def get_data(datadir):
    config.init(datadir)
    return config.BASEMENT.data



def get_settings(datadir):
    config.init(datadir)
    return config.BASEMENT.settings



#def get_params(datadir):
#    config.init(datadir)
#    return config.BASEMENT.data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 30 14:21:45 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
try:
    import emcee
except:
    pass



#::: translate the mcmc_move string into a list of emcee commands
def get_move(x):
    '''
    Parameters
    ----------
    x : str
        Denoting a single mcmc_move, e.g., 'StretchMove'.

    Raises
    ------
    ValueError
        Alerts the user if the input was corrupted.

    Returns
    -------
    emcee.moves objects
        The emcee.moves object responding to the input string.
    '''
    if x == 'RedBlueMove': 
        return emcee.moves.RedBlueMove()
    elif x == 'StretchMove': 
        return emcee.moves.StretchMove()
    elif x == 'WalkMove': 
        return emcee.moves.WalkMove()
    elif x == 'KDEMove': 
        return emcee.moves.KDEMove()
    elif x == 'DEMove': 
        return emcee.moves.DEMove()
    elif x == 'DESnookerMove': 
        return emcee.moves.DESnookerMove()
    elif x == 'MHMove': 
        return emcee.moves.MHMove()
    elif x == 'GaussianMove': 
        return emcee.moves.GaussianMove()
    else:
        raise ValueError('Acceptable values for the setting mcmc_move are: '+\
                         'RedBlueMove / StretchMove / WalkMove / KDEMove /'+\
                         'DEMove / DESnookerMove / MHMove / GaussianMove.'+\
                         'You may also give mixtures, e.g.,'+\
                         '"DEMove 0.8 DESnookerMove 0.2".'+\
                         'Here, however, "'+x+'" was given.')
            


def translate_str_to_move(mcmc_move_str):
    '''
    Converts 'mcmc_move' user input into a list of objects that emcee understands.

    Parameters
    ----------
    mcmc_move_str : string
        What the user inputs into the alelsfitter settings.csv file.
        E.g., 'StretchMove' 
        E.g., 'DEMove 0.8 DESnookerMove 0.2'

    Returns
    -------
    command_list : list
        What emcee expects for its 'move' argument.
        E.g., [(<emcee.moves.stretch.StretchMove(), 1.0)]
        E.g., [(<emcee.moves.de.DEMove(), 0.8), (<emcee.moves.de_snooker.DESnookerMove(), 0.2)]
    '''
    mcmc_move_list = mcmc_move_str.split(' ')
    if len(mcmc_move_list)%2!=0: 
        mcmc_move_list.append('1') # in case the user gives only one move (and thus no weights)
    object_list = [get_move(x) for x in mcmc_move_list[::2]]
    weight_list = [float(x) for x in mcmc_move_list[1::2]]
    command_list = [(x,y) for x,y in zip(object_list,weight_list)]
    return command_list
    

        
if __name__ == '__main__':
    '''
    For testing only.
    '''
    translate_str_to_move('StretchMove')
    translate_str_to_move('StretchMove 1')
    translate_str_to_move('DEMove 0.8 DESnookerMove 0.2')
    


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Credit & source: https://gist.github.com/salotz/4f585aac1adb6b14305c
"""


from __future__ import print_function, division, absolute_import

""" new colormaps from old: stack, truncate builtin cmaps / files / numpy arrays
What's a colormap or cmap in matplotlib ?
Mainly a bar or array of 256 colors, rgb or rgba values 0 .. 1,
used in
    pl.imshow( a 2d numpy array, cmap=cmap, ... )
    pl.colorbar()
Cmaps can be indexed with () like
    cmap( .25 ),  cmap( [0, .25, .5] ),  cmap( np.linspace( ... ))
to get rgb values.
The functions below return cmaps:
    get_cmap(): "Blues" ... builtins / filename / numpy array
    array_cmap(): a numpy array, n x 3 or 4  ints 0..255 or floats 0..1
    truncate_colormap(): subset
    stack_colormap(): A B -> bottom half, A, top half B.
    band_colormap(): e.g. 10 bands
See also
    http://matplotlib.org/api/colors_api.html   $matplotlib/colors.py
    http://matplotlib.org/api/cm_api.html       $matplotlib/cm.py
    http://en.wikipedia.org/wiki/Indexed_color
"""

import numpy as np
from matplotlib import pyplot as pl
from matplotlib import colors
from matplotlib.pyplot import colormaps

__version__ = "2013-12-19 dec denis"


def truncate_colormap( cmap, minval=0.0, maxval=1.0, n=256 ):
    """ mycolormap = truncate_colormap(
            cmap name or file or ndarray,
            minval=0.2, maxval=0.8 ): subset
            minval=1, maxval=0 )    : reverse
    by unutbu http://stackoverflow.com/questions/18926031/how-to-extract-a-subset-of-a-colormap-as-a-new-colormap-in-matplotlib
    """
    cmap = get_cmap( cmap )
    name = "%s-trunc-%.2g-%.2g" % (cmap.name, minval, maxval)
    return colors.LinearSegmentedColormap.from_list(
        name, cmap( np.linspace( minval, maxval, n )))

def stack_colormap( A, B, n=256 ):
    """ low half -> A colors, high half -> B colors """
    A = get_cmap( A )
    B = get_cmap( B )
    name = "%s-%s" % (A.name, B.name)
    lin = np.linspace( 0, 1, n )
    return array_cmap( np.vstack(( A(lin), B(lin) )), name, n=n )

def get_cmap( cmap, name=None, n=256 ):
    """ in: a name "Blues" "BuGn_r" ... of a builtin cmap (case-sensitive)
        or a filename, np.loadtxt() n x 3 or 4  ints 0..255 or floats 0..1
        or a cmap already
        or a numpy array.
        See http://wiki.scipy.org/Cookbook/Matplotlib/Show_colormaps
        or in IPython, pl.cm.<tab>
    """
    if isinstance( cmap, colors.Colormap ):
        return cmap
    if isinstance( cmap, str ):
        if cmap in colormaps():
            return pl.get_cmap( cmap )  # "Blues" ...
        A = np.loadtxt( cmap, delimiter=None )  # None: white space
        name = name or cmap.split("/")[-1] .split(".")[0]  # .../xx.csv -> xx
    else:
        A = cmap  # numpy array or array-like
    return array_cmap( A, name, n=n )

def array_cmap( A, name=None, n=256 ):
    """ numpy array -> a cmap, matplotlib.colors.Colormap
        n x 3 or 4  ints 0 .. 255 or floats 0 ..1
    """
    A = np.asanyarray( A )
    assert A.ndim == 2  and A.shape[1] in (3, 4), \
            "array must be n x 3 or 4, not %s" % str(A.shape)
    Amin, Amax = A.min(), A.max()
    if A.dtype.kind == "i":
        assert 0 <= Amin < Amax <= 255, "Amin %d  Amax %d must be in 0 .. 255" % (Amin, Amax)
        A = A / 255.  # not /=
    else:
        assert 0 <= Amin < Amax <= 1, "Amin %g  Amax %g must be in 0 .. 1" % (Amin, Amax)
    return colors.LinearSegmentedColormap.from_list( name or "noname", A, N=n )

def save_cmap( outfile, cmap ):
    """ -> a file of 256 x 4 ints 0 .. 255
        to load it, np.loadtxt() or get_cmap( filename )
    """
    cmap = get_cmap( cmap )
    A = cmap( np.linspace( 0, 1, 256 ))
    np.savetxt( outfile, A * 255, fmt="%4.0f",
            header="colormap %s" % cmap.name )  # From ...

def band_colormap( cmap, nband=10 ):
    """ -> a colormap with e.g. 10 bands """
    cmap = get_cmap( cmap )
    h = .5 / nband
    A = cmap( np.linspace( h, 1 - h, nband ))
    name = "%s-band-%d" % (cmap.name, nband)
    return array_cmap( A, name, n=nband )

#...............................................................................
cmap_brown = truncate_colormap( pl.cm.PuOr, minval=.5, maxval=0 )  # left half, flipped
cmap_bluebrown = stack_colormap( "Blues_r", cmap_brown )
cmap_bluebrown10 = band_colormap( cmap_bluebrown, 10 )
    # Tufte, Envisioning info p. 91: bathymetric


#...............................................................................
if __name__ == "__main__":
    import sys

    cmap = cmap_bluebrown10
    bw = array_cmap( [ [0.,0,0], [1,1,1] ], name="bw", n=2 )
    plot = 1

        # run this.py a=1 b=None c=[3] 'd = expr' ...  in sh or ipython
    exec( "\n".join( sys.argv[1:] ))
    np.set_printoptions( 2, threshold=100, edgeitems=10, linewidth=100, suppress=True )

    print(cmap.name, "\n", cmap( np.arange( 120, 136 ) / 256 ).T)
    save_cmap( cmap.name + ".tmp", cmap )

    if plot:
        A = np.arange(8**2) .reshape((8,8))
        im = pl.imshow( A, cmap=cmap, interpolation="nearest" )  # nearest: big squares
            # imshow_mouse_z
        pl.colorbar( im )
pl.show()



import os, csv
import numpy as np
from .. import config
from . import estimate_noise
from allesfitter.utils.latex_printer import round_txt_separately

def estimate_noise_wrap(pathdata):
    
    print('\nEstimating errors and baselines... this will take a couple of minutes. Please be patient, you will get notified once everything is completed.\n')
    
    #::: run MCMC fit to estimate errors and baselines
    estimate_noise.estimate_noise(pathdata)
    
    def fwrite_params(key, label, unit, physical_bounds):
        if INPUT[key+'_bounds_type'].value == 'uniform':
            bounds = 'uniform ' \
                     + str( np.max( [physical_bounds[0], float(INPUT[key+'_median'].value)-float(INPUT[key+'_lerr'].value)] ) ) + ' ' \
                     + str( np.min( [physical_bounds[1], float(INPUT[key+'_median'].value)+float(INPUT[key+'_uerr'].value)] ) )
        elif INPUT[key+'_bounds_type'].value == 'uniform * 5':
            bounds = 'uniform ' \
                     + str( np.max( [physical_bounds[0], float(INPUT[key+'_median'].value)-5*float(INPUT[key+'_lerr'].value)] ) ) + ' ' \
                     + str( np.min( [physical_bounds[1], float(INPUT[key+'_median'].value)+5*float(INPUT[key+'_uerr'].value)] ) )
        elif INPUT[key+'_bounds_type'].value == 'trunc_normal':
            bounds = 'trunc_normal ' \
                     + str(physical_bounds[0]) + ' ' \
                     + str(physical_bounds[1]) + ' ' \
                     + str(INPUT[key+'_median'].value) + ' ' \
                     + str(np.max( [ float(INPUT[key+'_lerr'].value), float(INPUT[key+'_uerr'].value) ] ))
        elif INPUT[key+'_bounds_type'].value == 'trunc_normal * 5':
            bounds = 'trunc_normal ' \
                     + str(physical_bounds[0]) + ' ' \
                     + str(physical_bounds[1]) + ' ' \
                     + str(INPUT[key+'_median'].value) + ' ' \
                     + str(5*np.max( [ float(INPUT[key+'_lerr'].value), float(INPUT[key+'_uerr'].value) ] ))
        fwrite(key + ',' + str(INPUT[key+'_median'].value) + ',' + str(int(INPUT[key+'_fit'].value)) + ',' +  bounds + ',' + label + ',' + unit)
    
    
    def fwrite(text):
        fname_params = os.path.join(pathdata, 'params.csv')
        with open(fname_params, 'a') as f:
            f.write(text+'\\n')
    
    
    def get_median_and_error_strings(text_median, text_lerr, text_uerr):
        if (text_median.value == ''):
            median = 'NaN'
            nan_fields = True
        else:
            median = text_median.value
        if (text_lerr.value == '') or (text_uerr.value == ''):
            err = 'NaN'
            nan_fields = True
        else:
            err = str( 5.* np.max( [float(text_lerr.value), float(text_uerr.value)] ) )
        median, err, _ = round_txt_separately( float(median), float(err), float(err) )
        return median, err
         
    
    def clean_up_csv(fname, N_last_rows=0):
        
        with open(fname, "r") as f:
            params_csv = list(csv.reader(f))
    
        with open(fname, "w") as f:
            writer = csv.writer(f)
            for i in range(len(params_csv)-N_last_rows):
                row = params_csv[i]
                writer.writerow(row)
    
    
    
    #::: delete the rows containing the default (zero) errors and baselines from the params.csv file
    N_default_rows = 0
    clean_up_csv( os.path.join( pathdata, 'params.csv' ), N_last_rows=N_default_rows )
    
    config.init(pathdata)
        
    #::: write new rows into params.csv
    #::: errors
    fwrite('#errors per instrument,')
    
    for i, inst in enumerate(config.BASEMENT.settings['inst_phot']):         
        #::: read in the summary file
        summaryfile = os.path.join( pathdata, 'priors', 'summary_phot.csv' )
        priors2 = np.genfromtxt(summaryfile, names=True, delimiter=',', dtype=None)
        priors = {}
        for key in priors2.dtype.names:
            priors[key] = np.atleast_1d(priors2[key])
    
        median = priors['ln_yerr_median'][i]
        err = 5.*np.max([ float(priors['ln_yerr_ll'][i]), float(priors['ln_yerr_ul'][i]) ])
        median, err, _ = round_txt_separately(median,err,err)
        fwrite('ln_err_flux_'+inst+','+median+',1,trunc_normal -23 0 '+median+' '+err+',$\log{\sigma_\mathrm{'+inst+'}}$,')
    
    for i, inst in enumerate(config.BASEMENT.settings['inst_rv']):   
        #::: read in the summary file
        summaryfile = os.path.join( pathdata, 'priors', 'summary_rv.csv' )
        priors2 = np.genfromtxt(summaryfile, names=True, delimiter=',', dtype=None)
        priors = {}
        for key in priors2.dtype.names:
            priors[key] = np.atleast_1d(priors2[key])
    
        median = priors['ln_yerr_median'][i]
        err = 5.*np.max([ float(priors['ln_yerr_ll'][i]), float(priors['ln_yerr_ul'][i]) ])
        median, err, _ = round_txt_separately(median,err,err)
        fwrite('ln_jitter_rv_'+inst+','+median+',1,trunc_normal -23 0 '+median+' '+err+',$\log{\sigma_\mathrm{jitter; '+inst+'}}$,')
    
    
    #::: write new rows into params.csv
    #::: baselines
    fwrite('#baseline per instrument,')
    
    for i, inst in enumerate(config.BASEMENT.settings['inst_phot']):         
        #::: read in the summary file
        summaryfile = os.path.join( pathdata, 'priors', 'summary_phot.csv' )
        priors2 = np.genfromtxt(summaryfile, names=True, delimiter=',', dtype=None)
        priors = {}
        for key in priors2.dtype.names:
            priors[key] = np.atleast_1d(priors2[key])
    
        median = priors['gp_ln_sigma_median'][i]
        err = 5.*np.max([ float(priors['gp_ln_sigma_ll'][i]), float(priors['gp_ln_sigma_ul'][i]) ])
        median, err, _ = round_txt_separately(median,err,err)
        fwrite('baseline_gp1_flux_'+inst+','+median+',1,trunc_normal -23 23 '+median+' '+err+',$\mathrm{gp: \log{\sigma} ('+inst+')}$,')
    
        median = priors['gp_ln_rho_median'][i]
        err = 5.*np.max([ float(priors['gp_ln_rho_ll'][i]), float(priors['gp_ln_rho_ul'][i]) ])
        median, err, _ = round_txt_separately(median,err,err)
        fwrite('baseline_gp2_flux_'+inst+','+median+',1,trunc_normal -23 23 '+median+' '+err+',$\mathrm{gp: \log{\\rho} ('+inst+')}$,')








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 18 12:48:58 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os
# import sys
import pathlib
import numpy as np
# from numpy.polynomial import Polynomial
# import matplotlib.pyplot as plt
import pandas as pd
# from tqdm import tqdm
# from glob import glob
# from pprint import pprint
from astropy.constants import G, M_earth, M_jup, M_sun, R_earth, R_jup, R_sun, au
from astropy import units as u
from time import time as timer

#::: my modules
# import allesfitter

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})


"""
Readme:
Astropy units are nice and fancy, but can cause a 17x slow-down for parts of this code.
Use constants' values instead of units.
"""


###############################################################################
#::: Convert Period P (in days) to semi-major axis a (in AU)
###############################################################################
def P_to_a(P, Mp, Ms):
    '''
    Parameters
    ----------
    P : float or array
        Planet orbital period, in days.
    Mp : float or array
        Planet mass, in Mearth.
    Ms : float or array
        Star mass, in Msun.

    Returns
    -------
    a : float or array
        Planet orbital semi-major axis, in AU.
    '''
    #apply units (convert to kg m s)
    P = P*86400. #in s
    Ms = Ms*M_sun.value #in kg
    Mp = Mp*M_earth.value #in kg
    
    return ( (G.value/(4*np.pi**2) * P**2 * np.cbrt(Ms + Mp)) ) / au.value  #in AU 



###############################################################################
#::: Same as above, and actual not much slower with astropy units
###############################################################################
def P_to_a_astropy(P, Mp, Ms):
    '''
    Parameters
    ----------
    P : float or array
        Planet orbital period, in days.
    Mp : float or array
        Planet mass, in Mearth.
    Ms : float or array
        Star mass, in Msun.

    Returns
    -------
    a : float or array
        Planet orbital semi-major axis, in AU.
    '''
    return ( (G/(4*np.pi**2) * (P*u.d)**2 * np.cbrt(Ms*u.Msun + Mp*u.Mearth)) ).to(u.AU).value  #in AU 



###############################################################################
#::: Convert semi-amplitude K to a companion mass
###############################################################################
def calc_M_comp_from_RV(K, P, incl, ecc, M_host, 
                        return_unit='M_earth',
                        approx=False):
    """
    This code can use the full binary mass function:
    M_comp**3 * (sini)**3 / (M_host + M_comp)**2 
     = (P * K**3) / (2. * pi * G) * (1. - e**2)**(3./2.)
    (no assumption that Mp << Ms needed).

    Parameters
    ----------
    K : float or array
        Host's RV semi-major amplitude.
    P : float or array
        Companion's orbital period, in days.
    incl : float or array
        Inclination, in degrees.
    ecc : float or array
        Eccentricity.
    M_host : float or array
        Host's mass, in Msun.

    Returns
    -------
    M_comp : float or array
        Companion's mass, in units of "return_unit".
    """
    #apply units
    P = P*86400. #in s
    K = K*1e3 #from km/s to m/s
    M_host = M_host*M_sun.value #in kg
    
    #return_unit
    if return_unit=='M_earth': return_unit = M_earth.value
    if return_unit=='M_jup': return_unit = M_jup.value
    if return_unit=='M_sun': return_unit = M_sun.value
    
    #chose a rough but sensible conversion unit for least awkward scaling
    if K < 1: 
        conv_unit = M_earth.value #in kg. #if it's less than m/s, chose Earth mass
    elif K < 1e3:
        conv_unit = M_jup.value #in kg. #if it's less than km/s, chose Jupiter mass
    else:
        conv_unit = M_sun.value #in kg. #if it's more, chose Sun mass
    
    #use the full binary mass function
    if not approx:
        
        #We rearange the binary mass function as a 3rd order polynomial.
        #We start with these definitions of a and b:
        a = np.sin(np.deg2rad(incl))**3 #dimensionless
        b = ( P * K**3 / (2.*np.pi*G.value) * (1. - ecc**2)**(3./2.) ) #in kg 
        #We use Mjup for the least awkward scaling, 
        #as this equation can be used for Earth-like exoplanets and binaries alike
        
        #Rearanging the binary mass function then gives us:
        #a * M_comp**3 - b * M_comp**2 - 2*b*M_host * M_comp - b*M_host**2 = 0
            
        #To use M_comp as a scalar for numpy, 
        #we also divide the whole equation by u.Mjup**3.
        #The coefficients have to absorb that.
        #(We again use Mjup for the least awkward scaling.)
        #This gives us:
        #a * M_comp**3 - b/u.Mjup * M_comp**2 - 2*b*M_host/u.Mjup**2 * M_comp - b*M_host**2/u.Mjup**3 = 0
        #Now divide by a to give the highest order term some independence
        #M_comp**3 - b/a/u.Mjup * M_comp**2 - 2*b/a*M_host/u.Mjup**2 * M_comp - b/a*M_host**2/u.Mjup**3 = 0
        
        
        #Now we define p3,...,p0 to make it even simpler:
        #p3 * M_comp**3 + p2 * M_comp**2 + p1 * M_comp + p0 = 0,
        #with the following coefficients:
        
        # p3 = 1.
        p2 = -b/a / conv_unit
        p1 = -2*b/a*(M_host) / (conv_unit**2)
        p0 = -b/a*(M_host)**2 / (conv_unit**3)
        # print(p2, p1, p0)
        
        #::: numpy's poly.roots() is too slow for us
        # poly = Polynomial([p0,p1,p2,p3])
        # s = poly.roots() #get all three roots; at least one is real
        # s = np.real_if_close(s[np.isreal(s)][0]) #in Mjup but unitless
        # s = (s*conv_unit).to(return_unit) #in return_unit
        # print(s)
        
        #::: instead, use some good old-fashioned high school math
        #::: see, e.g., http://teacherlink.ed.usu.edu/tlnasa/reference/ImagineDVD/Files/imagine/YBA/cyg-X1-mass/binary-formula.html#return
        #::: and http://teacherlink.ed.usu.edu/tlnasa/reference/ImagineDVD/Files/imagine/YBA/cyg-X1-mass/cubic.html
        Q = (3*p1 - p2**2)/9.
        R = (9.*p2*p1 - 27.*p0 - 2.*p2**3)/54. 
        
        S = np.cbrt(R + np.sqrt(Q**3 + R**2))
        T = np.cbrt(R - np.sqrt(Q**3 + R**2))
    
        s = S + T - p2/3. 
        s = (s*conv_unit) / return_unit
        
        return s
    
        
    #use the exoplanet approximation (just for testing)
    else:
        return ( K / np.sin(incl) * np.sqrt(1 - ecc**2) \
                 * np.cbrt(P / (2*np.pi*G) * (M_host)**2) ) / return_unit 
               #TODO, something is not right here
            


###############################################################################
#::: Same as above, but 17x slower with astropy units
###############################################################################            
def calc_M_comp_from_RV_astropy(K, P, incl, ecc, M_host, 
                        return_unit=u.Mearth,
                        approx=False):
    """
    This code can use the full binary mass function:
    M_comp**3 * (sini)**3 / (M_host + M_comp)**2 
     = (P * K**3) / (2. * pi * G) * (1. - e**2)**(3./2.)
    (no assumption that Mp << Ms needed).

    Parameters
    ----------
    K : float or array
        Host's RV semi-major amplitude.
    P : float or array
        Companion's orbital period, in days.
    incl : float or array
        Inclination, in degrees.
    ecc : float or array
        Eccentricity.
    M_host : float or array
        Host's mass, in Msun.

    Returns
    -------
    M_comp : float or array
        Companion's mass, in units of "return_unit".
    """
    #apply units
    P = P*u.d
    incl = incl*u.deg
    K = K*u.km/u.s
    
    #chose a rough but sensible conversion unit for least awkward scaling
    if K.value < 1e-3: 
        conv_unit = u.Mearth #if it's less than m/s, chose Earth mass
    elif K.value < 1:
        conv_unit = u.Mjup #if it's less than k/s, chose Jupiter mass
    else:
        conv_unit = u.Msun #if it's more, chose Sun mass
    
    #use the full binary mass function
    if not approx:
        
        #We rearange the binary mass function as a 3rd order polynomial.
        #We start with these definitions of a and b:
        a = np.sin(incl)**3 #dimensionless
        b = ( P * K**3 / (2.*np.pi*G) * (1. - ecc**2)**(3./2.) ).to(conv_unit) #in Mjup 
        #We use Mjup for the least awkward scaling, 
        #as this equation can be used for Earth-like exoplanets and binaries alike
        
        #Rearanging the binary mass function then gives us:
        #a * M_comp**3 - b * M_comp**2 - 2*b*M_host * M_comp - b*M_host**2 = 0
            
        #To use M_comp as a scalar for numpy, 
        #we also divide the whole equation by u.Mjup**3.
        #The coefficients have to absorb that.
        #(We again use Mjup for the least awkward scaling.)
        #This gives us:
        #a * M_comp**3 - b/u.Mjup * M_comp**2 - 2*b*M_host/u.Mjup**2 * M_comp - b*M_host**2/u.Mjup**3 = 0
        #Now divide by a to give the highest order term some independence
        #M_comp**3 - b/a/u.Mjup * M_comp**2 - 2*b/a*M_host/u.Mjup**2 * M_comp - b/a*M_host**2/u.Mjup**3 = 0
        
        
        #Now we define p3,...,p0 to make it even simpler:
        #p3 * M_comp**3 + p2 * M_comp**2 + p1 * M_comp + p0 = 0,
        #with the following coefficients:
        
        # p3 = 1.
        p2 = -b/a / conv_unit
        p1 = -2*b/a*(M_host*u.Msun) / (conv_unit**2)
        p0 = -b/a*(M_host*u.M_sun)**2 / (conv_unit**3)
        # print(p3, p2, p1, p0)
        
        #now the whole equation and all coefficients are dimensionless
        # p3 = p3.decompose()
        # p2 = p2.decompose()
        # p1 = p1.decompose()
        # p0 = p0.decompose()
        # print(p3, p2, p1, p0)
        
        #now the whole equation and all coefficients are dimensionless
        # p3 = p3.decompose().value
        p2 = p2.decompose().value 
        p1 = p1.decompose().value 
        p0 = p0.decompose().value
        # print(p3, p2, p1, p0)
        
        #::: numpy's poly.roots() is too slow for us
        # poly = Polynomial([p0,p1,p2,p3])
        # s = poly.roots() #get all three roots; at least one is real
        # s = np.real_if_close(s[np.isreal(s)][0]) #in Mjup but unitless
        # s = (s*conv_unit).to(return_unit) #in return_unit
        # print(s)
        
        #::: instead, use some good old-fashioned high school math
        Q = (3*p1 - p2**2)/9.
        R = (9.*p2*p1 - 27.*p0 - 2.*p2**3)/54. 
        
        S = np.cbrt(R + np.sqrt(Q**3 + R**2))
        T = np.cbrt(R - np.sqrt(Q**3 + R**2))
    
        s = S + T - p2/3. 
        s = (s*conv_unit).to(return_unit)
        
        return s.value
    
        
    #use the exoplanet approximation (just for testing)
    else:
        return ( K / np.sin(incl) \
                 * np.sqrt(1 - ecc**2) \
                 * np.cbrt(P / (2*np.pi*G) * (M_host)**2) \
               ).decompose().to(return_unit).value



###############################################################################
#::: Calculate a body's density
###############################################################################
def calc_rho(R, M, 
            R_unit = u.Rsun,
            M_unit = u.Msun,
            return_unit='cgs'):
    """
    Assumes a spherical body.
    
    Parameters
    ----------
    R : array or float
        Body's radius.
    M : array or float
        Body's mass.
    R_unit : astropy unit, optional
        Radius unit. The default is u.Rsun.
    M_unit : array or float, optional
        Mass unit. The default is u.Msun.
    return_unit : str, optional
        Return unit. The default is 'cgs'.

    Returns
    -------
    None.
    """
    #apply units
    R *= R_unit
    M *= M_unit
    
    #calculate
    V = 4./3. * np.pi * R**3
    rho = M / V
    
    #return
    if return_unit == 'cgs':
        return rho.cgs.value
    else:
        return None #TODO
    
    
    
###############################################################################
#::: Derive the host's density from transit and RV observations
###############################################################################
def calc_rho_host(P, radius_1, rr, rho_comp,
                           return_unit='cgs'):
    """
    Assumes a spherical body.
    
    Parameters
    ----------
    P : array or float
        Period, in days.
    radius_1 : array or float
        R_host / a.
    rr : astropy unit, optional
        R_comp / R_host.
    rho_comp : array or float, optional
        Density of the companion, in cgs units.
    return_unit : str, optional
        Return unit. The default is 'cgs'.

    Returns
    -------
    None.
    """
    #apply units
    P *= u.d #in days
    rho_comp = rho_comp * u.g / (u.cm)**3 #in cgs
    
    #calculate
    rho_host = ( (3*np.pi) / (G*P**2) * (1./radius_1)**3 - rr**3 * rho_comp ).decompose()
    
    #return
    if return_unit == 'cgs':
        return rho_host.cgs.value
    else:
        return None #TODO



###############################################################################
#::: Convert AU units to Rsun units
###############################################################################
def au_to_rsun(x): 
    '''
    Parameters
    ----------
    x : float or array
        Distance meassurement in AU.

    Returns
    -------
    x : float or array
        Distance meassurement in Rsun.
    '''
    return (np.array(x)*u.AU).to(u.Rsun).value



###############################################################################
#::: Estimate Kempton TSM from an exoplanet archive data frame
###############################################################################
def estimate_tsm_exoarchive(data):
    '''
    Parameters
    ----------
    data : dictionary
        must contain the following fields: 
        pl_rade, st_rad, sy_jmag, st_teff, pl_orbsmax, pl_masse

    Returns
    -------
    See estimate_tsm()
    '''
    
    Rp = data['pl_rade']
    Mp = data['pl_masse']
    a = data['pl_orbsmax']
    Rstar = data['st_rad']
    Teff = data['st_teff']
    mJ = data['sy_jmag']
    
    return estimate_tsm(Rp, Mp, a, Rstar, Teff, mJ)
    
    
    
###############################################################################
#::: Estimate Kempton TSM from individual variables
###############################################################################
def estimate_tsm(Rp, Mp, a, Rstar, Teff, mJ):
    '''
    Parameters
    ----------
    Rp : float or array
        Planet radius, in Rearth.
    Mp : float or array
        Planet mass, in Mearth.
    a : float or array
        Planet orbital semi-major axis, in AU.
    Rstar : float or array
        Star radius, in Rsun.
    Teff : float or array
        Star effective temperature, in Kelvin.
    mJ : float or array
        Star J-band magnitude.

    Returns
    -------
    tsm : float or array
        Transmission Spectroscopy Metric (TSM; via Kempton+ 2018).
    jwst_threshold : float or array
        Threshold for the TSM (via Kempton+ 2018).
        If the TSM value is below this threshold, it is deemed not worth observing with JWST.
    jwst_recommendation : bool or array
        True if tsm >= jwst_threshold.
        False if tsm < jwst_threshold.
    '''
    
    #::: Cast to 1d arrays and copy
    Rp = np.atleast_1d(Rp)
    Mp = np.atleast_1d(Mp)
    a = np.atleast_1d(a)
    Rstar = np.atleast_1d(Rstar)
    Teff = np.atleast_1d(Teff)
    mJ = np.atleast_1d(mJ)
    
    #::: Estimate Planet Equilibirium Temperature following Kempton et al., 2018
    Teq_mid = estimate_teq(a, Rstar, Teff)[0]
    
    #::: Estimate Planet Mass, where missing, following Chen & Kipping (2017) as implemented by Louie et al. (2018)
    #::: (as suggested by Kempton et al., 2018)
    Mp_estimate = 1*Mp
    ind = np.isnan(Mp) & (Rp<=1.23)
    Mp_estimate[ind] = 0.9718 * Rp[ind]**3.58
    ind = np.isnan(Mp) & (Rp>1.23) & (Rp<14.26)
    Mp_estimate[ind] = 1.436 * Rp[ind]**1.70
    
    #::: Estimate scale factors following Kempton et al. (2018)
    tsm_scale_factor = np.nan * Rp
    tsm_scale_factor[Rp < 1.5] = 0.190
    tsm_scale_factor[(Rp>1.5) & (Rp<=2.75)] = 1.26
    tsm_scale_factor[(Rp>2.75) & (Rp<=4)] = 1.28
    tsm_scale_factor[(Rp>4) & (Rp<=10)] = 1.15

    tsm = tsm_scale_factor * Rp**3 * Teq_mid / Mp_estimate / Rstar**2 * 10**(-mJ/5.)
    
    return tsm, tsm_scale_factor, Teq_mid, Mp_estimate



###############################################################################
#::: Estimate planet equilibrium temperature
###############################################################################
def estimate_teq(a, Rstar, Teff):
    '''
    Parameters
    ----------
    a : float or array
        Planet orbital semi-major axis, in AU.
    Teff : float or array
        Star effective temperature, in Kelvin.
    Rstar : float or array
        Star radius, in Rsun.

    Returns
    -------
    Teq_mid : float or array
        Middle-value of the planet equilibrium temperature,
        estimated for zero albedo (A=0) and full day-night heat redistribution (E=1).
    Teq_low : float or array
        Low value of the planet equilibrium temperature,
        estimated for albedo A=0.3 and day-night heat redistribution E=1.
    Teq_high : float or array
        High value of the planet equilibrium temperature,
        estimated for albedo A=0 and day-night heat redistribution E=1,
        if the atmosphere instantaneously reradiates the absorbed radiation (with no advection), f = 2/3 (e.g., Seager 2010).
    '''
    
    #a) the middle Teq
    #Estimate Teq for zero albedo (A=0) and full day-night heat redistribution (E=1)
    #[suggested by Kempton et al. (2018)]
    Teq_mid = (Teff * np.sqrt(Rstar / au_to_rsun(a)) * ( 1/4. )**(0.25))
     
    #b) the minimum Teq
    #Estimate Teq for albedo A=0.3 and day-night heat redistribution E=1
    #like for TOI-270, like for Earth
    Teq_low = (Teff * np.sqrt(Rstar / au_to_rsun(a)) * ( (1/4.) * (1-0.3) )**(0.25))
     
    #b) the maximum Teq
    #Estimate Teq for albedo A=0 and day-night heat redistribution E=1
    #If the atmosphere instantaneously reradiates the absorbed radiation (with no advection), f = 2/3 (e.g., Seager 2010).
    Teq_high = (Teff * np.sqrt(Rstar / au_to_rsun(a)) * ( 2/3. )**(0.25))
    
    return Teq_mid, Teq_low, Teq_high



###############################################################################
#::: Estimate TTV super-period of 1st order mean-motion-resonance (MMR)
###############################################################################
def estimate_ttv_super_period_of_first_order_mmr(P1, P2, MMR='2:1'):
    '''
    Estimates the TTV super-period.
    Only works for first order MMRs, e.g., 2:1, 3:2, 4:3, etc.
    Following Eq. 7 of Lithwick+ 2017, https://iopscience.iop.org/article/10.1088/0004-637X/761/2/122/pdf
    
    Parameters
    ----------
    P1 : float
        Orbital period of the inner planet.
    P2 : float
        Orbital period of the outer planet.
    MMR : str, optional
        Mean motion resonance. 
        The larger number must come first.
        The default is '2:1'.

    Returns
    -------
    TTV super-period : float
        The TTV super-period.
    '''
    
    j = int(MMR.split(':')[0])
    return 1. / np.abs( (1.*j/P2) - (1.*(j-1.)/P1) )




###############################################################################
#::: Estimate tidal locking time scale
###############################################################################
#::: THIS ONE DIDN'T WORK
def estimate_tidal_locking_time_scale_Wikipedia(P, R_companion, M_companion, M_host, typ='rock'):
    if typ == 'rock': mu = 3e10
    elif typ == 'ice': mu = 4e9
    elif typ == 'gas': mu = 4e9
    
    a = P_to_a(P, M_companion, M_host) # in AU
    a = (a*u.AU).to(u.m).value # in m
    R_companion = (R_companion*u.Rearth).to(u.m).value # in m
    M_companion = (M_companion*u.Mearth).to(u.kg).value # in kg
    M_host = (M_host*u.Msun).to(u.kg).value # in kg
    
    return 6 * (a**6 * R_companion * mu) / (M_companion + M_host**2) * 1e10



def estimate_tidal_locking_time_scale_Kastings1993(P, M_host):
    '''
    This assumes an Earth-like planet (1 Rearth, 1 Mearth, rocky).

    Parameters
    ----------
    P : float or array
        Planet orbital period, in days.
    M_host : float or array
        Mass of the host, in Msun.

    Returns
    -------
    tlock : float
        tidal locking time scale, in years.
    '''
    
    a = P_to_a(P, 1., M_host) # in AU
    a = (a*u.AU).to(u.cm).value # in cm
    M_host = (M_host*u.Msun).to(u.g).value # in grams
    
    return (1./486. * ( a / (0.027 * M_host**(1./3.)) )**6 *u.s).to(u.yr).value # in years




###############################################################################
#::: Estimate spectral types
###############################################################################
def estimate_spectral_type(Teff):
    """
    Estimate the spectral type of a main-sequence dwarf star from the Teff only.
    No check for dwarfity, giantness, or any other stuff.
    Contains: 03-9.5, B0-9.5, A0-9.5, F0-9.5, G0-9.5, K0-9.5, M0-9.5, L0-9.5, T0-9.5, Y0-2
    
    Parameters
    ----------
    Teff : float or list of float
        Stellar effective temperature
        For example [3300, 4400, np.nan]
        
    Returns
    -------
    SpT : str or list of str
        For example: ['M3.5V', 'K5V', None] 
    """
    
    here = pathlib.Path(__file__).parent.absolute()
    f = os.path.join(here,'_static','_stars','Peacut_Mamajek.csv')
    df = pd.read_csv(f)[::-1].reset_index(drop=True) # sort by ascending Teff
    Teff = np.atleast_1d(Teff) # ensure it is iterable

    if len(Teff)==1 and np.isnan(Teff[0]):
        return None
    
    elif len(Teff)==1: 
        return df['#SpT'][ np.argmin( np.abs(df['Teff'] - Teff[0]) ) ].value
        
    else:
        Teff[ (Teff<=200) | (Teff>=50000) | (np.isnan(Teff)) ] = 1e6 # replace trash with a large number
        bins = [200] + list(df['Teff'][:-1] + np.diff(df['Teff'])/2.) + [50000] # set bins, length of df+1
        ind = np.digitize(Teff, bins) - 1 # -1 because digitize counts as bins[i-1] <= x < bins[i]
        SpT_all = np.append( df['#SpT'], None ) # length of df+1, last entry is for trash
        return np.array( SpT_all[ ind ] )
    
    
    
def estimate_spectral_class(Teff):
    """
    Estimates the spectral class. 
    Similar to estimate_spectral_type(Teff), but with larger bins.
    Contains: 0, B, A, F, G, K, early M, mid M, late M, L, T, Y
    early M : M0-M3.5
    mid M : M4-M6.5
    late M : M7-M9.5
    
    Parameters
    ----------
    Teff : float or list of float
        Stellar effective temperature
        For example [3300, 4400, np.nan]

    Returns
    -------
    SpC : str or list of str
        For example: ['early M', 'K', None] 
    """
    
    SpT = estimate_spectral_type(Teff)
    SpC = []
    for s in np.atleast_1d(SpT):
        if s is None:
            SpC.append(None)
        elif s[0]=='M':
            x = float(s[1:-1])
            if (x>=0) and (x<=3.5):
                SpC.append('early M')
            elif (x>=4) and (x<=6.5):
                SpC.append('mid M')
            elif (x>=7) and (x<=9.5):
                SpC.append('late M')
        else:
            SpC.append(s[0])
            
    if len(SpC) == 1:
        return SpC[0]
    else:
        return np.array( SpC )
    
    
    
##########################################################################
#::: MAIN
##########################################################################
if __name__ == '__main__':
    
    ##########################################################################
    #::: speed test
    ##########################################################################
    N = 1000
    
    #::: 
    t0 = timer()
    for i in range(N):
        calc_M_comp_from_RV(K = 13e-3,P = 12*365, incl = 90 ,ecc = 0, M_host = 1, return_unit = u.Mjup, approx=False)
    t1 = timer()
    dt12 = (t1-t0)/N
    print('Runtime with floats:', dt12*1e3, 'ms')
    
    t3 = timer()
    for i in range(N):
        calc_M_comp_from_RV_astropy(K = 13e-3,P = 12*365, incl = 90 ,ecc = 0, M_host = 1, return_unit = u.Mjup, approx=False)
    t4 = timer()
    dt34 = (t4-t3)/N
    print('Runtime with astropy units:',  dt34*1e3, 'ms')

    print('Runntime speed up by using floats instead of astropy units:',  dt34/dt12)
    
    
    ##########################################################################
    #::: accuracy test
    ##########################################################################
    jup_exact = calc_M_comp_from_RV(K = 13e-3, 
                                    P = 12*365, 
                                    incl = 90, 
                                    ecc = 0, 
                                    M_host = 1, 
                                    return_unit = u.Mjup,
                                    approx=False)
    print('Jupiter Mass, exact:', jup_exact, 'Mjup')
    
    
    # jup_approx = calc_M_comp_from_RV(K = 13e-3, 
    #                                 P = 12*365, 
    #                                 incl = 90, 
    #                                 ecc = 0, 
    #                                 M_host = 1, 
    #                                 return_unit = u.Mjup,
    #                                 approx=True)
    # print('Jupiter Mass, approx:', jup_approx)
    
    
    ear_exact = calc_M_comp_from_RV(K = 9e-5, 
                                    P = 365, 
                                    incl = 90, 
                                    ecc = 0, 
                                    M_host = 1, 
                                    return_unit = u.Mearth,
                                    approx=False)
    print('Earth Mass, exact:', ear_exact, 'Mearth')
    
    
    # ear_approx = calc_M_comp_from_RV(K = 13e-3, 
    #                                 P = 12*365, 
    #                                 incl = 90, 
    #                                 ecc = 0, 
    #                                 M_host = 1, 
    #                                 return_unit = u.Mjup,
    #                                 approx=False)
    # print('Earth Mass, exact:', ear_approx)
    
    
    ##########################################################################
    #::: speed test
    ##########################################################################
    N = 10000
    
    #::: 
    t0 = timer()
    for i in range(N):
        P_to_a(1,1,1)
    t1 = timer()
    dt12 = (t1-t0)/N
    print('P_to_a: Runtime with floats:', dt12*1e3, 'ms')
    
    t3 = timer()
    for i in range(N):
        P_to_a(1,1,1)
    t4 = timer()
    dt34 = (t4-t3)/N
    print('P_to_a: Runtime with astropy units:',  dt34*1e3, 'ms')

    print('P_to_a: Runntime speed up by using floats instead of astropy units:',  dt34/dt12)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 10 18:27:01 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import pandas as pd
from tqdm import tqdm
from glob import glob
from pprint import pprint
try: 
    from brokenaxes import brokenaxes
except: 
    pass #don't throw an error for now, otherwise people will panic
import pathlib
from astropy.time import Time

#::: local imports
from allesfitter.io import read_csv
from allesfitter.time_series import sigma_clip, slide_clip

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in", "ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth':1})

#::: globals (nice)
TJD_offset = 2457000

    
    

def clean_up(time, y, yerr=None, time_format='BJD_TDB'):
    time = np.array(time)    
    y = np.array(y)    
    if yerr is not None: 
        yerr = np.array(yerr)    
    
    if time_format == 'BJD_TDB': 
        time = time - TJD_offset
    
    #TODO: the below messes things up in case a "trend" is passed, too
    #TODO: it is probably not even needed, so just remove it?
    # if yerr is None:
    #     ind = np.where( ~np.isnan(time*y) )
    #     time = time[ind]
    #     y = y[ind]    
    # else:
    #     ind = np.where( ~np.isnan(time*y*yerr) )
    #     time = time[ind]
    #     y = y[ind]    
    #     yerr = yerr[ind]
        
    return time, y, yerr




def guess_labels(ax, time, y):
    
    if (np.nanpercentile(y,5) > 0.7) & (np.nanpercentile(y,95) < 1.3):
        ax.set_ylabel('Flux')
    elif (np.nanpercentile(y,5) > -0.02) & (np.nanpercentile(y,95) < 0.02):
        ax.set_ylabel('Residuals')
    else:
        ax.set_ylabel('RV (km/s)')
       
    ax.set_xlabel(r'Time (BJD$_\mathrm{TDB}$ - '+str(TJD_offset)+')')




def fullplot(time, y, yerr=None, ax=None, time_format='BJD_TDB', clip=False):
    '''
    Parameters
    ----------
    time : array of float
        e.g. time array (usually in days)
    y : array of float
        e.g. flux or RV array (usually as normalized flux or RV in km/s)
    yerr : array of float
        e.g. flux or RV error array (usually as normalized flux or RV in km/s)
    ax : TYPE, optional
        DESCRIPTION. The default is None.
    time_format : str
        The format of your time array. 
        Must be either 'BJD_TDB' or 'TJD' (TESS Julian Date). 
        The default is 'BJD_TDB'.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.

    Returns
    -------
    ax : TYPE
        DESCRIPTION.
    '''
    
    if ax is None: fig, ax = plt.subplots(figsize=(12,3), tight_layout=True)
    time, y, yerr = clean_up(time, y, yerr, time_format)
    if clip:
        flux, mask_lower, mask_upper = _clip_helper(time, y)
    
    ax.errorbar(time, y, yerr=yerr, fmt='b.', ms=2, rasterized=True)
    if clip:
        ax.plot(time*mask_upper, ax.get_ylim()[1]*mask_upper, 'r^', color='orange', ms=10, zorder=11)
        ax.plot(time*mask_lower, ax.get_ylim()[0]*mask_lower, 'rv', color='orange', ms=10, zorder=11)
    guess_labels(ax, time, y)

    return ax



def fullplot_csv(fname, ax=None, time_format='BJD_TDB'):
    '''
    Wrapper around fullplot to plot straight from a csv file.
    See fullplot() for details.
    '''
    time, y, yerr = read_csv(fname)[0:3]
    return fullplot(time, y, yerr, ax=ax, time_format=time_format)



def brokenplot(time, y, yerr=None, trend=None, dt=10, ax=None, time_format='BJD_TDB', fmt='b.', clip=False):
    '''
    Parameters
    ----------
    time : array of float
        e.g. time array (usually in days)
    y : array of float
        e.g. flux or RV array (usually as normalized flux or RV in km/s)
    yerr : array of float
        e.g. flux or RV error array (usually as normalized flux or RV in km/s)
    dt : float, optional
        The gap size after which axes will be broken. The default is 10 (usually in days).
    ax : TYPE, optional
        DESCRIPTION. The default is None.
    time_format : str
        The format of your time array. 
        Must be either 'BJD_TDB' or 'TJD' (TESS Julian Date). 
        The default is 'BJD_TDB'.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.

    Returns
    -------
    bax : brokenaxes instance
        Just like an pyplot.Axes instance
    '''
    
    if 'brokenaxes' not in sys.modules:
        raise ImportError('You must install the brokenaxes package before using allesfitter.brokenplot()\n(pip install brokenaxes)')
    
    if ax is None: fig, ax = plt.subplots(figsize=(12,3))
    time, y, yerr = clean_up(time, y, yerr, time_format)
    if clip:
        flux, mask_lower, mask_upper = _clip_helper(time, y)
        
    ind0 = [0] + list(np.where(np.diff(time)>dt)[0]+1) #start indices of data chunks
    ind1 = list(np.where(np.diff(time)>dt)[0]) + [len(time)-1] #end indices of data chunks
    xlims = [ (time[i]-(time[j]-time[i])/100.,time[j]+(time[j]-time[i])/100.) for i,j in zip(ind0,ind1) ]
    
    ax.set_axis_off() #empty the axis before brokenaxes does its magic
    bax = brokenaxes(xlims=xlims, subplot_spec=ax.get_subplotspec())
    bax.errorbar(time, y, yerr=yerr, fmt=fmt, ms=2, rasterized=True)
    
    if trend is not None:
        bax.plot(time, trend, 'r-', lw=2, zorder=100)
    
    if clip:
        bax.plot(time*mask_upper, bax.axs[0].get_ylim()[1]*mask_upper, 'r^', color='orange', ms=10, zorder=11)
        bax.plot(time*mask_lower, bax.axs[0].get_ylim()[0]*mask_lower, 'rv', color='orange', ms=10, zorder=11)
    # bax.ticklabel_format(axis='y', style='sci', useOffset=True)
    # plt.gca().yaxis.set_major_locator(plt.MaxNLocator(3))
    guess_labels(bax, time, y)
    bax.set_ylabel('Flux\n')
    
    return bax



def brokenplot_csv(fname, dt=10, ax=None, time_format='BJD_TDB'):
    '''
    Wrapper around brokenplot to plot straight from a csv file.
    See brokenplot() for details.
    '''
    time, y, yerr = read_csv(fname)[0:3]
    return brokenplot(time, y, yerr, dt=dt, ax=ax, time_format=time_format)



def tessplot(time, flux, flux_err=None, trend=None, time_format='BJD_TDB', clip=False, sharey=False, axes=None, shade=True, **kwargs):
    '''
    Creates a new line for every new TESS Sector
    
    Parameters
    ----------
    time : array of float
        Time stamps (in days).
    flux : array of float
        Flux.
    flux_err : array of float, optional
        Flux error bars. The default is None.
    time_format : str
        The format of your time array.  Must be either 'BJD_TDB' or 
        'TJD' (TESS Julian Date). The default is 'BJD_TDB'.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.
    sharey : bool
        Share the y-axis between all rows. The default is True.
    kwargs : ...
        Any keyword arguments you wish to pass to the plot function (e.g., color='red')

    Returns
    -------
    None.
    '''
    
    time, flux, flux_err = clean_up(time, flux, flux_err, time_format)
    if clip:
        flux, mask_lower, mask_upper = _clip_helper(time, flux)
        
    here = pathlib.Path(__file__).parent.absolute()
    df = pd.read_csv( os.path.join(here,'_static','_tess','tess_orbit_times_by_sector.csv'), skiprows=5)
    
    inds, sectors = [], []
    for s in range(1,max(df['Sector'])):
        line = 2*(s-1)
        t0 = float(df['Start TJD'].loc[line])
        t1 = float(df['End TJD'].loc[line+1])
        ind = np.where( (time > t0) & (time <= t1) )[0]
        if len(ind)>0:
            inds += [list(ind)]
            sectors += [s]
    
    N = len(inds)
    # ss = ax.get_subplotspec()
    # ax.set_axis_off() #empty the axis before brokenaxes does its magic
    # gssub = ss.subgridspec(N,1)
    
    if axes is None:
        fig, axes = plt.subplots(N, figsize=(12,3*N), tight_layout=True, sharey=sharey)
    
    for i in range(N):
        # ax1 = fig.add_subplot(gssub[i,0])
        ax1 = np.atleast_1d(axes)[i]
        ind, s = inds[i], sectors[i]
        line = 2*(s-1)
        if flux_err is None:
            ax1.plot(time[ind], flux[ind], 'b.', ms=2, rasterized=True, **kwargs)
        else:
            ax1.errorbar(time[ind], flux[ind], yerr=flux_err[ind], fmt='b.', ms=2, rasterized=True, **kwargs)
        if clip:
            ax1.plot(time[ind]*mask_upper[ind], ax1.get_ylim()[1]*mask_upper[ind], 'r^', color='orange', ms=10, zorder=11)
            ax1.plot(time[ind]*mask_lower[ind], ax1.get_ylim()[0]*mask_lower[ind], 'rv', color='orange', ms=10, zorder=11)
        if trend is not None:
            ax1.plot(time[ind], trend[ind], 'r-', lw=2, rasterized=True)
        ax1.text(0.98,0.95,'Sector '+str(s),ha='right',va='top',transform=ax1.transAxes)
        orb0 = float(df['Start TJD'].loc[line])
        orb1 = float(df['End TJD'].loc[line])
        orb2 = float(df['Start TJD'].loc[line+1])
        orb3 = float(df['End TJD'].loc[line+1])
        # ax1.plot([orb0,orb1],[1,1],'k-',c='silver',lw=2)
        # ax1.plot([orb2,orb3],[1,1],'k-',c='silver',lw=2)
        if shade:
            ax1.axvspan(orb0,orb1,color='b',alpha=0.1,zorder=-11)
            ax1.axvspan(orb2,orb3,color='b',alpha=0.1,zorder=-11)
        ax1.set(xlim=[orb0-0.5, orb3+0.5], ylabel='Flux')
        ax1.yaxis.set_major_locator(plt.MaxNLocator(3))
        if i==N-1: ax1.set(xlabel=r'Time (BJD$_\mathrm{TDB}$ - '+str(TJD_offset)+')')
    
    return axes
    


def tessplot_csv(fname, time_format='BJD_TDB', **kwargs):
    '''
    Wrapper around tessplot to plot straight from a csv file.
    See tessplot() for details.
    '''
    time, flux, flux_err = read_csv(fname)[0:3]
    return tessplot(time, flux, flux_err, time_format=time_format, **kwargs)    



def monthplot(time, flux, flux_err=None, trend=None, clip=False, sharey=False, axes=None, **kwargs):
    '''
    Creates a new line for every calendar month
    
    Parameters
    ----------
    time : array of float
        Time stamps (in days).
    flux : array of float
        Flux.
    flux_err : array of float, optional
        Flux error bars. The default is None.
    time_format : str
        The format of your time array.  Must be either 'BJD_TDB' or 
        'TJD' (TESS Julian Date). The default is 'BJD_TDB'.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.
    sharey : bool
        Share the y-axis between all rows. The default is True.
    kwargs : ...
        Any keyword arguments you wish to pass to the plot function (e.g., color='red')

    Returns
    -------
    None.
    '''
    time, flux, flux_err = clean_up(time, flux, flux_err, time_format=None)
    if clip:
        flux, mask_lower, mask_upper = _clip_helper(time, flux)
    
    #::: get the months
    start_of_month = [str(year)+'-'+str(month).zfill(2)+'-01T00:00:00' 
                  for year in range(1950,2050) for month in range(1,12)]
    start_of_month = Time(start_of_month, format='isot', scale='utc')
    start_of_month_str = [t.strftime('%Y %b') for t in start_of_month]

    inds, start_jds, end_jds, months = [], [], [], []
    for i in range(len(start_of_month)-1):
        t0 = start_of_month.jd[i]
        t1 = start_of_month.jd[i+1]-1e-12
        ind = np.where( (time > t0) & (time <= t1) )[0]
        if len(ind)>0:
            inds.append( list(ind) )
            start_jds.append( start_of_month.jd[i] )
            end_jds.append( start_of_month.jd[i+1]-1e-12 )
            months.append( start_of_month_str[i] )
    
    N = len(inds)
    
    if axes is None:
        fig, axes = plt.subplots(N, figsize=(12,3*N), tight_layout=True, sharey=sharey)
    
    for i in range(N):
        # ax1 = fig.add_subplot(gssub[i,0])
        ax1 = np.atleast_1d(axes)[i]
        ind, start, end, month = inds[i], start_jds[i], end_jds[i], months[i]
        # line = 2*(s-1)
        if flux_err is None:
            ax1.plot(time[ind], flux[ind], 'b.', ms=2, rasterized=True, **kwargs)
        else:
            ax1.errorbar(time[ind], flux[ind], yerr=flux_err[ind], fmt='b.', ms=2, rasterized=True, **kwargs)
        if clip:
            ax1.plot(time[ind]*mask_upper[ind], ax1.get_ylim()[1]*mask_upper[ind], 'r^', color='orange', ms=10, zorder=11)
            ax1.plot(time[ind]*mask_lower[ind], ax1.get_ylim()[0]*mask_lower[ind], 'rv', color='orange', ms=10, zorder=11)
        if trend is not None:
            ax1.plot(time[ind], trend[ind], 'r-', lw=2, rasterized=True)
        ax1.text(0.98,0.95,month,ha='right',va='top',transform=ax1.transAxes)
        # orb0 = float(df['Start TJD'].loc[line])
        # orb1 = float(df['End TJD'].loc[line])
        # orb2 = float(df['Start TJD'].loc[line+1])
        # orb3 = float(df['End TJD'].loc[line+1])
        # ax1.plot([orb0,orb1],[1,1],'k-',c='silver',lw=2)
        # ax1.plot([orb2,orb3],[1,1],'k-',c='silver',lw=2)
        # if shade:
        #     ax1.axvspan(orb0,orb1,color='b',alpha=0.1,zorder=-11)
        #     ax1.axvspan(orb2,orb3,color='b',alpha=0.1,zorder=-11)
        ax1.set(xlim=[start-0.5, end+0.5], ylabel='Flux')
        ax1.yaxis.set_major_locator(plt.MaxNLocator(3))
        if i==N-1: ax1.set(xlabel=r'Time (BJD$_\mathrm{TDB}$ - '+str(TJD_offset)+')')
    
    return fig
    


def monthplot_csv(fname, **kwargs):
    '''
    Wrapper around tessplot to plot straight from a csv file.
    See tessplot() for details.
    '''
    time, flux, flux_err = read_csv(fname)[0:3]
    return monthplot(time, flux, flux_err, **kwargs)    



def _clip_helper(time, y):
    y, _, mask_upper0, mask_lower0 = sigma_clip(time, y, low=4, high=4, return_mask=True)
    y, _, mask_upper, mask_lower = slide_clip(time, y, window_length=1, low=4, high=4, return_mask=True)
    mask_upper = np.array(mask_upper0*mask_upper, dtype=float)
    mask_upper[mask_upper==0] = np.nan
    mask_lower = np.array(mask_lower0*mask_lower, dtype=float)
    mask_lower[mask_lower==0] = np.nan
    
    return y, mask_lower, mask_upper


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 16 17:15:26 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
# import os
# import sys
import numpy as np
# import matplotlib.pyplot as plt
# import pandas as pd
# from tqdm import tqdm
# from glob import glob
# from pprint import pprint
from astropy.stats import sigma_clip as sigma_clip_

#::: specific modules
try:
    from wotan import flatten
except ImportError:
    pass

#::: my modules
from .exoworlds_rdx.lightcurves.lightcurve_tools import rebin_err

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




###############################################################################
#::: clean a time series
###############################################################################
def clean(time, y, y_err=None):
    """
    Cleans all input from masks, NaN or Inf values and returns them as np.ndarray.
    Careful: This changes the length of the arrays!

    Parameters
    ----------
    time : array of flaot
        Time stamps.
    y : array of float
        Any data corresponding to the time stamps.
    y_err : array of float or None, optional
        Error of the data. The default is None.

    Returns
    -------
    Cleaned time, y, and y_err (shorter arrays!)
    
    Explanation
    -----------
    The input might be of a mixutre of types, such as list, np.ndarray, or np.ma.core.MaskedArray.
    For stability, first convert them all to np.ma.core.MaskedArray.
    This will preserve all information, even non-masked NaN and Inf values.
    Then mask all invalid values, retrieve the mask, and use it to remove those values.
    Finally, convert them to a np.ndarray.
    Tested on the following example:
        input:
            time = np.linspace(1,6,6)
            flux = np.ma.array([1,2,3,4,np.nan,np.inf], 
                               mask=[True,True,False,False,False,False])
            flux_err = None
        returns:
            (array([3., 4.]), array([3., 4.]), None)
    """

    time = np.ma.array(time)
    y = np.ma.array(y)
    if y_err is None:
        mask = np.ma.masked_invalid(time*y).mask
    else:
        y_err = np.ma.array(y_err)
        mask = np.ma.masked_invalid(time*y*y_err).mask
        y_err = np.array(y_err[~mask])
    time = np.array(time[~mask])
    y = np.array(y[~mask])
        
    return time, y, y_err
        


###############################################################################
#::: sort a time series
###############################################################################
def sort(time, y, y_err=None):
    """
    Sorts all input in time; the input should be cleaned first.

    Parameters
    ----------
    time : array of flaot
        Time stamps.
    y : array of float
        Any data corresponding to the time stamps.
    y_err : array of float or None, optional
        Error of the data. The default is None.

    Returns
    -------
    Sorted time, y, and y_err.
    """
    ind_sort = np.argsort(time)
    time = time[ind_sort]
    y = y[ind_sort]
    if y_err is not None:
        y_err = y_err[ind_sort]
        
    return time, y, y_err



###############################################################################
#::: slide clip a time series
###############################################################################
def sigma_clip(time, y, low=4, high=4, return_mask=False):
    """
    Astropy's sigma_clip but returning NaN instead of a masked array.

    Parameters
    ----------
    time : array of flaot
        Time stamps.
    y : array of float
        Any data corresponding to the time stamps.
    low : float, optional
        The lower sigma. The default is 5.
    high : float, optional
        The upper sigma. The default is 5.
    return_mask : bool, optional
        Return the masks or only the clipped time series. The default is False.

    Returns
    -------
    Clipped y (outliers replaced with NaN).
    """
    y2 = sigma_clip_(np.ma.masked_invalid(y), sigma_lower=low, sigma_upper=high) #astropy wants masked arrays
    mask = y2.mask
    y2 = np.array(y2.filled(np.nan)) #use NaN instead of masked arrays, because masked arrays drive me crazy
    
    if not return_mask:
        return y2
    
    else: 
        with np.testing.suppress_warnings() as sup:
            sup.filter(UserWarning)
            mask_upper = (y > np.nanmedian(y2)) * mask
            mask_lower = (y < np.nanmedian(y2)) * mask
        return y2, mask, mask_upper, mask_lower



###############################################################################
#::: slide clip a time series
###############################################################################
def slide_clip(time, y, window_length=1, low=4, high=4, return_mask=False):
    """
    Slide clip outliers from a non-stationary time series;
    a much faster alternative to Wotan's built-in slide clip.

    Parameters
    ----------
    time : array of flaot
        Time stamps.
    y : array of float
        Any data corresponding to the time stamps.\
    window_length : float, optional
        The length of the sliding window (in time's units). The default is 1.
    low : float, optional
        The lower sigma. The default is 5.
    high : float, optional
        The upper sigma. The default is 5.
    return_mask : bool, optional
        Return the masks or only the clipped time series. The default is False.

    Returns
    -------
    Clipped y (outliers replaced with NaN).
    """
    y_flat = flatten(time, y, method='biweight', window_length=window_length)
    y2, mask, mask_upper, mask_lower = sigma_clip(time, y_flat, low=low, high=high, return_mask=True)
    y3 = 1*y
    y3[mask] = np.nan
    
    if not return_mask:
        return y3
    
    else:
        return y3, mask, mask_upper, mask_lower
    
    

###############################################################################
#::: binning
###############################################################################
def binning(time, y, y_err=None, dt=None):
    """
    Bin a time series to bin-widths specified by dt.
    This method also handles gaps and error bars properly.
    Careful: This will change the length of the arrays!

    Parameters
    ----------
    time : array of flaot
        Time stamps.
    y : array of float
        Any data corresponding to the time stamps.
    y_err : array of float or None, optional
        Error of the data. The default is None.
    dt : float or None, optional
        Bin-widths in time's units. The default is None (no binning).

    Returns
    -------
    Binned time, y, and y_err (shorter arrays!).
    """
    time, y, y_err = clean(time, y, y_err)
    if dt is not None:
        return rebin_err(time, y, y_err, dt=dt, 
                         ferr_type='medsig', ferr_style='sem', sigmaclip=True)[0:3]
    else:
        return time, y, y_err



###############################################################################
#::: mask bad data regions
###############################################################################
def mask_regions(time, y, bad_regions=None):
    """
    Mask regions by filling y and y_err with NaN for those selected times.
    
    Parameters
    ----------
    time : array of flaot
        Time stamps.
    y : array of float
        Any data corresponding to the time stamps.
    y_err : array of float or None, optional
        Error of the data. The default is None.
    bad_regions : list or None, optional
        List of tuples like [(start0,end0),(start1,end1),...], 
        where any (start,end) are the start and end points of bad data bad_regions. 
        The default is None.

    Returns
    -------
    Masked time, y, and y_err (bad_regions of y and y_err replaced with NaN)
    """
    y2 = 1.*y
    
    if bad_regions is not None:
        for bad_region in bad_regions:
            ind_bad = np.where((time>=bad_region[0]) & (time<=bad_region[1]))[0]
            y2[ind_bad] = np.nan
    
    return y2


'''
Hold the analytic flare model
Could also put other simple flare models in here, e.g. polynomial from Balona (2015)
Algorithm originally from here: https://github.com/jradavenport/FBEYE
'''

import numpy as np
from scipy.stats import binned_statistic

def aflare(t, p):
    """
    This is the Analytic Flare Model from the flare-morphology paper.
    Reference Davenport et al. (2014) http://arxiv.org/abs/1411.3723
    Note: this model assumes the flux before the flare is zero centered
    Note: many sub-flares can be modeled by this method by changing the
    number of parameters in "p". As a result, this routine may not work
    for fitting with methods like scipy.optimize.curve_fit, which require
    a fixed number of free parameters. Instead, for fitting a single peak
    use the aflare1 method.
    Parameters
    ----------
    t : 1-d array
        The time array to evaluate the flare over
    p : 1-d array
        p == [tpeak, fwhm (units of time), amplitude (units of flux)] x N
    Returns
    -------
    flare : 1-d array
        The flux of the flare model evaluated at each time
    """
    _fr = [1.00000, 1.94053, -0.175084, -2.24588, -1.12498]
    _fd = [0.689008, -1.60053, 0.302963, -0.278318]

    Nflare = int( np.floor( (len(p)/3.0) ) )

    flare = np.zeros_like(t)
    # compute the flare model for each flare
    for i in range(Nflare):
        outm = np.piecewise(t, [(t<= p[0+i*3]) * (t-p[0+i*3])/p[1+i*3] > -1.,
                                (t > p[0+i*3])],
                            [lambda x: (_fr[0]+                             # 0th order
                                        _fr[1]*((x-p[0+i*3])/p[1+i*3])+     # 1st order
                                        _fr[2]*((x-p[0+i*3])/p[1+i*3])**2.+  # 2nd order
                                        _fr[3]*((x-p[0+i*3])/p[1+i*3])**3.+  # 3rd order
                                        _fr[4]*((x-p[0+i*3])/p[1+i*3])**4. ),# 4th order
                             lambda x: (_fd[0]*np.exp( ((x-p[0+i*3])/p[1+i*3])*_fd[1] ) +
                                        _fd[2]*np.exp( ((x-p[0+i*3])/p[1+i*3])*_fd[3] ))]
                            ) * p[2+i*3] # amplitude
        flare = flare + outm
        
    return flare


def aflare1(t, tpeak, fwhm, ampl, upsample=False, uptime=10):
    '''
    The Analytic Flare Model evaluated for a single-peak (classical).
    Reference Davenport et al. (2014) http://arxiv.org/abs/1411.3723
    Use this function for fitting classical flares with most curve_fit
    tools.
    Note: this model assumes the flux before the flare is zero centered
    Parameters
    ----------
    t : 1-d array
        The time array to evaluate the flare over
    tpeak : float
        The time of the flare peak
    fwhm : float
        The "Full Width at Half Maximum", timescale of the flare
    ampl : float
        The amplitude of the flare
    upsample : bool
        If True up-sample the model flare to ensure more precise energies.
    uptime : float
        How many times to up-sample the data (Default is 10)
    Returns
    -------
    flare : 1-d array
        The flux of the flare model evaluated at each time
    '''
    _fr = [1.00000, 1.94053, -0.175084, -2.24588, -1.12498]
    _fd = [0.689008, -1.60053, 0.302963, -0.278318]

    if upsample:
        dt = np.nanmedian(np.diff(t))
        timeup = np.linspace(min(t)-dt, max(t)+dt, t.size * uptime)

        flareup = np.piecewise(timeup, [(timeup<= tpeak) * (timeup-tpeak)/fwhm > -1.,
                                        (timeup > tpeak)],
                                    [lambda x: (_fr[0]+                       # 0th order
                                                _fr[1]*((x-tpeak)/fwhm)+      # 1st order
                                                _fr[2]*((x-tpeak)/fwhm)**2.+  # 2nd order
                                                _fr[3]*((x-tpeak)/fwhm)**3.+  # 3rd order
                                                _fr[4]*((x-tpeak)/fwhm)**4. ),# 4th order
                                     lambda x: (_fd[0]*np.exp( ((x-tpeak)/fwhm)*_fd[1] ) +
                                                _fd[2]*np.exp( ((x-tpeak)/fwhm)*_fd[3] ))]
                                    ) * np.abs(ampl) # amplitude

        # and now downsample back to the original time...
        ## this way might be better, but makes assumption of uniform time bins
        # flare = np.nanmean(flareup.reshape(-1, uptime), axis=1)

        ## This way does linear interp. back to any input time grid
        # flare = np.interp(t, timeup, flareup)

        ## this was uses "binned statistic"
        downbins = np.concatenate((t-dt/2.,[max(t)+dt/2.]))
        flare,_,_ = binned_statistic(timeup, flareup, statistic='mean',
                                 bins=downbins)

    else:
        flare = np.piecewise(t, [(t<= tpeak) * (t-tpeak)/fwhm > -1.,
                                 (t > tpeak)],
                                [lambda x: (_fr[0]+                       # 0th order
                                            _fr[1]*((x-tpeak)/fwhm)+      # 1st order
                                            _fr[2]*((x-tpeak)/fwhm)**2.+  # 2nd order
                                            _fr[3]*((x-tpeak)/fwhm)**3.+  # 3rd order
                                            _fr[4]*((x-tpeak)/fwhm)**4. ),# 4th order
                                 lambda x: (_fd[0]*np.exp( ((x-tpeak)/fwhm)*_fd[1] ) +
                                            _fd[2]*np.exp( ((x-tpeak)/fwhm)*_fd[3] ))]
                                ) * np.abs(ampl) # amplitude

    return flare


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 19:00:41 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm
from glob import glob
from pprint import pprint
from scipy.stats import anderson
from statsmodels.tsa.stattools import adfuller
from statsmodels.stats.stattools import durbin_watson 
from statsmodels.stats.diagnostic import acorr_ljungbox


#::: allesfitter modules
from . import config
from .general_output import logprint

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




def alles_anderson(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.

    Returns
    -------
    isNormal : bool
        True if the residuals are normally distributed, False otherwise.
    
    Outputs
    -------
    It also prints the statstics and conclusions.
    
    Sauces
    ------
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson.html
    https://www.statology.org/anderson-darling-test-python/
    '''
    logprint('Anderson-Darling Test')
    logprint('---------------------')
    logprint('This tests the null hypothesis that the residuals follows a normal distribution.') 
    statistic, critical_values, significance_levels = anderson(residuals)
    logprint('Test statistic\t\t', statistic)
    logprint('Critical values\t\t', critical_values)
    logprint('Significance levels\t', significance_levels/100.)
    logprint('Does the null hypotheses hold at a significance level of...')
    for cv, sl in zip(critical_values, significance_levels):
        logprint('...', sl/100., '\t\t', statistic < cv)
    isNormal = all(statistic < critical_values)
    if isNormal:
        logprint('The null hypothesis cannot be rejected.')
        logprint('In simple words: your residuals look good.')
    else:
        logprint('The null hypothesis is rejected at some significance levels.')
        logprint('In simple words: there might still be some structure in your residuals.')
    logprint('\n')
    
    return isNormal
        


def alles_adfuller(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.

    Returns
    -------
    isStationary : bool
        True if the residuals are stationary, False otherwise.
    
    Outputs
    -------
    It also prints the statstics and conclusions.
    
    Sauces
    ------
    https://www.hackdeploy.com/augmented-dickey-fuller-test-in-python/
    '''
    class StationarityTests:
        def __init__(self, significance=.05):
            self.SignificanceLevel = significance
            self.pValue = None
            self.isStationary = None
            
        def ADF_Stationarity_Test(self, timeseries, printResults = True):
            logprint('Augmented Dickey-Fuller Test')
            logprint('----------------------------')
            logprint('This tests the null hypothesis that the residuals show non-stationarity (trends).') 
            adfTest = adfuller(timeseries, autolag='AIC')
            self.pValue = adfTest[1]
            if (self.pValue<self.SignificanceLevel):
                self.isStationary = True
            else:
                self.isStationary = False
            if printResults:
                dfResults = pd.Series(adfTest[0:4], index=['Test Statistic','P-Value','# Lags Used','# Observations Used'])
                for key,value in adfTest[4].items():
                    dfResults['Critical Value (%s)'%key] = value
                logprint(dfResults.to_string())
            
    sTest = StationarityTests()
    sTest.ADF_Stationarity_Test(residuals, printResults = True)
    logprint("Is the time series stationary? {0}".format(sTest.isStationary))
    if not sTest.isStationary:
        logprint('In simple words: there might still be some structure in your residuals.')
    else:
        logprint('In simple words: your residuals look good.')
    logprint('\n')
    
    return sTest.isStationary
        
    

def alles_durbin(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.

    Returns
    -------
    isUncorrelated : bool
        True if the residuals are not correlated, False otherwise.
    
    Outputs
    -------
    It also prints the statstics and conclusions.
    
    Sauces
    ------
    https://www.statology.org/durbin-watson-test-python/
    '''
    statistic = durbin_watson(residuals) 
    isUncorrelated = (statistic>=1.5) & (statistic<=2.5)
    logprint('Durbin-Watson Test')
    logprint('------------------')
    logprint('This tests the null hypothesis that there is no correlation among the residuals.')
    logprint('Test statistic:', statistic)
    logprint('The test statistic is approximately equal to 2*(1-r) where r is the sample autocorrelation of the residuals. Interpretation:')
    logprint('\t< 1.5 suggests a positive correlation')
    logprint('\t1.5 to 2.5 suggests no correlation')
    logprint('\t> 2.5 suggests a negative correlation')
    if isUncorrelated:
        logprint('No signs of a correlation.')
        logprint('In simple words: your residuals look good.')
    elif statistic<1.5:
        logprint('Signs of a positive correlation.')
        logprint('In simple words: there might still be some structure in your residuals.')
    elif statistic>2.5:
        logprint('Signs of a negative correlation.')
        logprint('In simple words: there might still be some structure in your residuals.')
    logprint('\n')

    return isUncorrelated 

    

def alles_ljung(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.

    Returns
    -------
    isUncorrelated : bool
        True if the residuals are not correlated, False otherwise.
    
    Outputs
    -------
    It also prints the statstics and conclusions.
    
    Sauces
    ------
    https://www.statology.org/ljung-box-test-python/
    https://www.statsmodels.org/dev/generated/statsmodels.stats.diagnostic.acorr_ljungbox.html?highlight=ljung
    '''
    
    logprint('Ljung-Box Test')
    logprint('--------------')
    logprint('This tests the null hypothesis that there is no correlation among the residuals.')
    
    df = acorr_ljungbox(residuals, lags=[1,5,10,15,20], return_df=True)
    df.reset_index(inplace=True)
    df = df.rename(columns = {'index':'lag'})
    logprint('Does the null hypotheses hold at a significance level of...')
    df['0.15'] = df['lb_pvalue'] > 0.15 #if Ture, the null hypothesis cannot be rejected at this significance level
    df['0.1'] = df['lb_pvalue'] > 0.10 #if Ture, the null hypothesis cannot be rejected at this significance level
    df['0.05'] = df['lb_pvalue'] > 0.05 #if Ture, the null hypothesis cannot be rejected at this significance level
    df['0.025'] = df['lb_pvalue'] > 0.025 #if Ture, the null hypothesis cannot be rejected at this significance level
    df['0.01'] = df['lb_pvalue'] > 0.01 #if Ture, the null hypothesis cannot be rejected at this significance level
    isUncorrelated = all(df['0.15']==True) & all(df['0.1']==True) & all(df['0.05']==True) & all(df['0.025']==True) & all(df['0.01']==True)
    logprint(df.to_string(index=False))
    if isUncorrelated:
        logprint('The null hypothesis cannot be rejected.')
        logprint('In simple words: your residuals look good.')
    else:
        logprint('The null hypothesis is rejected at some significance levels.')
        logprint('In simple words: there might still be some structure in your residuals.')
    logprint('\n')

    return isUncorrelated 
    
    
    
def residual_stats(residuals):
    '''
    Parameters
    ----------
    residuals : array of float
        The residuals after the fit of model+baseline+stellar_var.
    typ : str
        'mcmc' or 'ns', just givs the name to the output file.

    Returns
    -------
    None.
    
    Outputs
    -------
    Prints the statstics and conclusions.
    '''
    logprint("\nPerforming diagnostic tests on the fit's residuals...\n")
    passed_anderson = alles_anderson(residuals)
    passed_adfuller = alles_adfuller(residuals)
    passed_durbin = alles_durbin(residuals)
    passed_ljung = alles_ljung(residuals)
    logprint('Summary')
    logprint('-------')
    logprint('Test                    Passed?')
    logprint('Anderson-Darling       ',passed_anderson)
    logprint('Augmented Dickey-Fuller',passed_adfuller)
    logprint('Durbin-Watson          ',passed_durbin)
    logprint('Ljung-Box              ',passed_ljung)
    
    
    
# if __name__ == "__main__":
#     residuals = np.random.normal(size=1000)
#     alles_residual_stats(residuals)


# ==============================================================================
# =============================     LDC3 v1.0     ==============================
# ==============================================================================
#
# AUTHOR: David Kipping
#         Columbia University, Dept. of Astronomy
#         Please report any problems to: d.kipping@columbia.edu
#
# CITATION: If using this code, please cite:
#           Kipping, D. M., 2015, 'Efficient, uninformative sampling of limb 
#           darkening coefficients for a three-parameter law', MNRAS, accepted
#
# DESCRIPTION: LDC3 is a module containing three subroutines: "forward", 
#              "inverse" and "criteriatest". These subroutines perform tasks
#              related to identifying physically plausible limb darkening 
#              coefficients (LDCs) in the case of the 3-parameter limb
#              darkening law proposed by Sing et al. (2009), A&A, 505, 891:
#              I(mu)/I(1) = 1 - c_2*[1-mu] - c_3*[1-mu^(3/2)] - c_4*[1-mu^2].
#              The expressions used in LDC3 were derived in Kipping (2015),
#              and we recommend the reader review this paper before continuing.
#              See each subroutine for a description of its purpose. Note that
#              alpha_t denotes alpha_theta throughout this code.
#
# HISTORY: v1.0 Initial version released

import math

#===============================================================================
def forward(alphas):

  # DESCRIPTION: forward takes a set of LDCs in the alpha-parameterization and
  #              converts them to the c-parameterization. A likely common use
  #              of this subroutine would be when fitting LDCs. One would treat
  #              the alpha-parameters as the free parameters, but then convert 
  #              them to the more standard c-parameters before calling whatever 
  #              astronomical model describes one's observations.
  #
  # INPUTS: alphas=[alpha_h, alpha_r, alpha_t], all of which are bound between 0 
  #         and 1 and the latter is a wrap-around parameter.
  #
  # OUTPUTS: c = [c_2, c_3, c_4]

  # Forward modeling constants
  third	= 0.333333333333333
  twopi	= 6.283185307179586
  P1	= 4.500841772313891
  P2	= 17.14213562373095
  Q1	= 7.996825477806030
  Q2	= 8.566161603278331

  # Define alphas
  alpha_h = alphas[0]
  alpha_r = alphas[1]
  alpha_t = alphas[2]

  # Compute LDCs
  c_2 = (alpha_h**third)*( P1 + 0.25*math.sqrt(alpha_r)*( \
        -6.0*math.cos(twopi*alpha_t) + P2*math.sin(twopi*alpha_t) ) )
  c_3 = (alpha_h**third)*( -Q1 - Q2*math.sqrt(alpha_r)*math.sin(twopi*alpha_t) )
  c_4 = (alpha_h**third)*( P1 + 0.25*math.sqrt(alpha_r)*( \
        6.0*math.cos(twopi*alpha_t) + P2*math.sin(twopi*alpha_t) ) )

  # Define c results
  c=[0 for i in range(3)]
  c[0] = c_2
  c[1] = c_3
  c[2] = c_4
  return c
#===============================================================================

#===============================================================================
def inverse(c):

  # DESCRIPTION: inverse performs the reverse transformation of subroutine 
  #              forward. It therefore converts c-parameters into 
  #              alpha-parameters.
  #
  # INPUTS: c = [c_2, c_3, c_4] (= LDCs)
  #
  # OUTPUTS: alphas = [alpha_h, alpha_r, alpha_t]

  # Inverse modeling constants
  third	= 0.3333333333333333
  twopi	= 6.2831853071795865
  F1	= 0.9997221357486548
  F2	= 1.0002947195445628
  G1	= 51.396969619669996
  G2	= 51.426406871192850
  G3	= 0.5098605589396862
  G4	= 67.195959492893320
  G5	= 75.639610306789280
  H1	= 0.4666386075895370
  H2	= 0.5252750715749255

  # Define c
  c_2 = c[0]
  c_3 = c[1]
  c_4 = c[2]

  # Compute inverse LDCs
  alpha_h = ( F1*c_2 + F2*c_3 + F1*c_4 )**3
  alpha_r = G3/( G1*c_2 + G2*c_3 + G1*c_4 )**2
  alpha_r = alpha_r*( 576.0*(c_2 - c_4)**2 + ( G4*c_2 + G5*c_3 + G4*c_4 )**2 )
  alpha_t = math.atan2( -H1*c_2 - H2*c_3 - H1*c_4 , third*0.5*(c_4-c_2) )
  alpha_t = ( alpha_t - twopi*math.floor( alpha_t/twopi ) )/twopi

  # Define alphas_inv results
  alphas_inv=[0 for i in range(3)]
  alphas_inv[0] = alpha_h
  alphas_inv[1] = alpha_r
  alphas_inv[2] = alpha_t
  return alphas_inv
#===============================================================================

#===============================================================================
def criteriatest(usemod,c):

  # DESCRIPTION: criteriatest tests whether a set of LDCs (with the 
  #              c-parameterization) satisfy the seven analytic criteria defined
  #              in Kipping (2015). Passing this test indicates that LDCs
  #              correspond to a physically allowed limb darkened intensity
  #              profile, as defined in Kipping (2015). One may choose to use
  #              modified or unmodified versions of the criteria with the usemod
  #              logical control. The modified versions are slightly more
  #              conservative, cropping ~5% of the allowed parameter volume, but
  #              yielding a more symmetric volume.
  #
  # INPUTS: usemod (integer; 1 = use modified criteria; 0 = use unmodified 
  #                          criteria)
  #         c = [c_2, c_3, c_4] (= LDCs)
  #
  # OUTPUTS: passed (integer; 1 = test passed, 0 = test failed)

  # Define c
  c_2 = c[0]
  c_3 = c[1]
  c_4 = c[2]
  passed = 1

  # Criteria A
  if (c_2+c_3+c_4)>1.0:
    passed = 0

  # Criteria B
  if (2.0*c_2+3.0*c_3+4.0*c_4)<0.0:
    passed = 0

  # Criteria C
  if c_2<0.0:
    passed = 0

  # Criteria D
  if (c_2+c_3+c_4)<0.0:
    passed = 0

  # Criteria E
  if c_3>0.0:
    passed = 0

  # Criteria F
  if usemod==1:
    # Modified Criterion F
    if c_4<0.0:
      passed = 0
  else:
    # Unmodified Criterion F
    if c_4<-1.0:
      passed = 0

  # Criteria G
  if usemod==1:
    # Modified Criterion G
    if (32.0*c_2*c_4)<(9.0*c_3*c_3):
      passed = 0
  else:
    # Unmodified Criterion G
    if (0.375*c_3/c_4)<=0.0 and (0.375*c_3/c_4)>=-1.0:
      if (32.0*c_2*c_4)<(9.0*c_3*c_3):
        passed = 0

  # Return the result
  return passed
#===============================================================================


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 23 11:40:38 2019

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from setuptools import setup, find_packages





setup(
    name = 'allesfitter',
    packages = find_packages(),
    version = '1.2.1',
    description = 'A global inference framework for photometry and RV',
    author = 'Maximilian N. Günther & Tansu Daylan',
    author_email = 'maxgue@mit.edu',
    url = 'https://github.com/MNGuenther/allesfitter',
    download_url = 'https://github.com/MNGuenther/allesfitter',
    license='MIT',
    classifiers=['Development Status :: 5 - Production/Stable', #3 - Alpha / 4 - Beta / 5 - Production/Stable
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python'],
    #install_requires=['numpy>=1.10'],
    include_package_data = True
    )



